# Copyright (C) 2013-2016 DNAnexus, Inc.
#
# This file is part of dx-toolkit (DNAnexus platform client libraries).
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may not
#   use this file except in compliance with the License. You may obtain a copy
#   of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

"""
DXApplet Handler
++++++++++++++++

Applets are data objects that store application logic, including
specifications for executing it, and (optionally) input and output
signatures. They can be run by calling the :func:`DXApplet.run` method.

"""

from __future__ import print_function, unicode_literals, division, absolute_import

import dxpy
from . import DXDataObject, DXJob
from ..utils import merge
from ..utils.resolver import is_project_id
from ..system_requirements import SystemRequirementsDict
from ..exceptions import DXError
from ..compat import basestring

class DXExecutable:
    '''Methods in :class:`!DXExecutable` are used by
    :class:`~dxpy.bindings.dxapp.DXApp`,
    :class:`~dxpy.bindings.dxapplet.DXApplet`,
    :class:`~dxpy.bindings.dxworkflow.DXWorkflow`, and
    :class:`~dxpy.bindings.dxworkflow.DXGlobalWorkflow`
    '''
    def __init__(self, *args, **kwargs):
        raise NotImplementedError("This class is a mix-in. Use DXApp or DXApplet instead.")

    @staticmethod
    def _get_run_input_common_fields(executable_input, **kwargs):
        '''
        Takes the same arguments as the run method. Creates an input hash for the /executable-xxxx/run method,
        translating ONLY the fields that can be handled uniformly across all executables: project, folder, name, tags,
        properties, details, depends_on, allow_ssh, debug, delay_workspace_destruction, ignore_reuse, and extra_args.
        '''
        project = kwargs.get('project') or dxpy.WORKSPACE_ID

        run_input = {"input": executable_input}
        for arg in ['folder', 'name', 'tags', 'properties', 'details']:
            if kwargs.get(arg) is not None:
                run_input[arg] = kwargs[arg]

        if any(kwargs.get(key) is not None for key in ['instance_type', 'cluster_spec', 'fpga_driver', 'nvidia_driver']):
            instance_type_srd = SystemRequirementsDict.from_instance_type(kwargs.get('instance_type'))
            cluster_spec_srd = SystemRequirementsDict(kwargs.get('cluster_spec'))
            fpga_driver_srd = SystemRequirementsDict(kwargs.get('fpga_driver'))
            nvidia_driver_srd = SystemRequirementsDict(kwargs.get('nvidia_driver'))
            run_input["systemRequirements"] = (instance_type_srd + cluster_spec_srd + fpga_driver_srd + nvidia_driver_srd).as_dict()

        if kwargs.get('system_requirements') is not None:
            run_input["systemRequirements"] = kwargs.get('system_requirements')

        if kwargs.get('system_requirements_by_executable') is not None:
            run_input["systemRequirementsByExecutable"] = kwargs.get('system_requirements_by_executable')

        if kwargs.get('depends_on') is not None:
            run_input["dependsOn"] = []
            if isinstance(kwargs['depends_on'], list):
                for item in kwargs['depends_on']:
                    if isinstance(item, DXJob) or isinstance(item, DXDataObject):
                        if item.get_id() is None:
                            raise DXError('A dxpy handler given in depends_on does not have an ID set')
                        run_input["dependsOn"].append(item.get_id())
                    elif isinstance(item, basestring):
                        run_input['dependsOn'].append(item)
                    else:
                        raise DXError('Expected elements of depends_on to only be either instances of DXJob or DXDataObject, or strings')
            else:
                raise DXError('Expected depends_on field to be a list')

        if kwargs.get('delay_workspace_destruction') is not None:
            run_input["delayWorkspaceDestruction"] = kwargs['delay_workspace_destruction']

        if kwargs.get('allow_ssh') is not None:
            run_input["allowSSH"] = kwargs['allow_ssh']

        if kwargs.get('debug') is not None:
            run_input["debug"] = kwargs['debug']

        if kwargs.get('priority') is not None:
            run_input["priority"] = kwargs['priority']

        if kwargs.get('ignore_reuse') is not None:
            run_input["ignoreReuse"] = kwargs['ignore_reuse']

        if dxpy.JOB_ID is None or (kwargs.get('detach') is True and project is not None and is_project_id(project)):
            run_input["project"] = project

        if kwargs.get('extra_args') is not None:
            merge(run_input, kwargs['extra_args'])

        if kwargs.get('detach') is not None:
            run_input["detach"] = kwargs['detach']

        if kwargs.get('cost_limit') is not None:
            run_input["costLimit"] = kwargs['cost_limit']

        if kwargs.get('rank') is not None:
            run_input["rank"] = kwargs['rank']

        if kwargs.get('max_tree_spot_wait_time') is not None:
            run_input["maxTreeSpotWaitTime"] = kwargs['max_tree_spot_wait_time']

        if kwargs.get('max_job_spot_wait_time') is not None:
            run_input["maxJobSpotWaitTime"] = kwargs['max_job_spot_wait_time']
        
        if kwargs.get('detailed_job_metrics') is not None:
            run_input["detailedJobMetrics"] = kwargs['detailed_job_metrics']

        preserve_job_outputs = kwargs.get('preserve_job_outputs')
        if preserve_job_outputs is not None and preserve_job_outputs != False:
            run_input["preserveJobOutputs"] = {} if preserve_job_outputs == True else preserve_job_outputs

        return run_input

    @staticmethod
    def _get_run_input_fields_for_applet(executable_input, **kwargs):
        '''
        Takes the same arguments as the run method. Creates an input
        hash for the /applet-xxxx/run method.
        '''
        # Although it says "for_applet", this is factored out of
        # DXApplet because apps currently use the same mechanism
        for unsupported_arg in ['stage_instance_types', 'stage_folders', 'rerun_stages', 'ignore_reuse_stages']:
            if kwargs.get(unsupported_arg):
                raise DXError(unsupported_arg + ' is not supported for applets (only workflows)')

        run_input = DXExecutable._get_run_input_common_fields(executable_input, **kwargs)

        if kwargs.get('head_job_on_demand') is not None:
            run_input["headJobOnDemand"] = kwargs['head_job_on_demand']
        return run_input

    def _run_impl(self, run_input, **kwargs):
        """
        Runs the executable with the specified input and returns a
        handler for the resulting execution object
        (:class:`~dxpy.bindings.dxjob.DXJob` or
        :class:`~dxpy.bindings.dxanalysis.DXAnalysis`).

        Any kwargs are passed on to :func:`~dxpy.DXHTTPRequest`.
        """
        raise NotImplementedError('_run_impl is not implemented')

    def _get_run_input(self, executable_input, **kwargs):
        """
        Takes the same arguments as the run method. Creates an input
        hash for the /executable-xxxx/run method.
        """
        raise NotImplementedError('_get_run_input is not implemented')

    def _get_required_keys(self):
        """
        Abstract method used in executable_unbuilder.dump_executable
        """
        raise NotImplementedError('_get_required_keys is not implemented')

    def _get_optional_keys(self):
        """
        Abstract method used in executable_unbuilder.dump_executable
        """
        raise NotImplementedError('_get_optional_keys is not implemented')

    def _get_describe_output_keys(self):
        """
        Abstract method used in executable_unbuilder.dump_executable
        """
        raise NotImplementedError('_get_describe_output_keys is not implemented')

    def _get_cleanup_keys(self):
        """
        Abstract method used in executable_unbuilder.dump_executable
        """
        raise NotImplementedError('_get_cleanup_keys is not implemented')

    def run(self, executable_input, project=None, folder=None, name=None, tags=None, properties=None, details=None,
            instance_type=None, stage_instance_types=None, stage_folders=None, rerun_stages=None, cluster_spec=None,
            depends_on=None, allow_ssh=None, debug=None, delay_workspace_destruction=None, priority=None, head_job_on_demand=None,
            ignore_reuse=None, ignore_reuse_stages=None, detach=None, cost_limit=None, rank=None, max_tree_spot_wait_time=None,
            max_job_spot_wait_time=None, preserve_job_outputs=None, detailed_job_metrics=None, extra_args=None,
            fpga_driver=None, system_requirements=None, system_requirements_by_executable=None, nvidia_driver=None, **kwargs):
        '''
        :param executable_input: Hash of the executable's input arguments
        :type executable_input: dict
        :param project: Project ID of the project context
        :type project: string
        :param folder: Folder in which executable's outputs will be placed in *project*
        :type folder: string
        :param name: Name for the new job (default is "<name of the executable>")
        :type name: string
        :param tags: Tags to associate with the job
        :type tags: list of strings
        :param properties: Properties to associate with the job
        :type properties: dict with string values
        :param details: Details to set for the job
        :type details: dict or list
        :param instance_type: Instance type on which the jobs will be run, or a dict mapping function names to instance type requests
        :type instance_type: string or dict
        :param depends_on: List of data objects or jobs to wait that need to enter the "closed" or "done" states, respectively, before the new job will be run; each element in the list can either be a dxpy handler or a string ID
        :type depends_on: list
        :param allow_ssh: List of hostname or IP masks to allow SSH connections from
        :type allow_ssh: list
        :param debug: Configuration options for job debugging
        :type debug: dict
        :param delay_workspace_destruction: Whether to keep the job's temporary workspace around for debugging purposes for 3 days after it succeeds or fails
        :type delay_workspace_destruction: boolean
        :param priority: Priority level to request for all jobs created in the execution tree, "low", "normal", or "high"
        :type priority: string
        :param head_job_on_demand: If true, the job will be run on a demand instance.
        :type head_job_on_demand: bool
        :param ignore_reuse: Disable job reuse for this execution
        :type ignore_reuse: boolean
        :param ignore_reuse_stages: Stages of a workflow (IDs, names, or indices) or "*" for which job reuse should be disabled
        :type ignore_reuse_stages: list
        :param detach: If provided, job will not start as subjob if run inside of a different job.
        :type detach: boolean
        :param cost_limit: Maximum cost of the job before termination.
        :type cost_limit: float
        :param rank: Rank of execution
        :type rank: int
        :param max_tree_spot_wait_time: Number of seconds allocated to each path in the root execution's tree to wait for Spot
        :type max_tree_spot_wait_time: int
        :param max_job_spot_wait_time: Number of seconds allocated to each job in the root execution's tree to wait for Spot
        :type max_job_spot_wait_time: int
        :param preserve_job_outputs: Copy cloneable outputs of every non-reused job entering "done" state in this root execution to a folder in the project. If value is True it will place job outputs into the "intermediateJobOutputs" subfolder under the output folder for the root execution. If the value is dict, it may contains "folder" key with desired folder path. If the folder path starts with '/' it refers to an absolute path within the project, otherwise, it refers to a subfolder under root execution's output folder.
        :type preserve_job_outputs: boolean or dict
        :param detailed_job_metrics: Enable detailed job metrics for this root execution
        :type detailed_job_metrics: boolean
        :param extra_args: If provided, a hash of options that will be merged into the underlying JSON given for the API call
        :type extra_args: dict
        :returns: Object handler of the newly created job
        :param fpga_driver: a dict mapping function names to fpga driver requests
        :type fpga_driver: dict
        :param system_requirements: System requirement single mapping
        :type system_requirements: dict
        :param system_requirements_by_executable: System requirement by executable double mapping
        :type system_requirements_by_executable: dict
        :param nvidia_driver: a dict mapping function names to nvidia driver requests
        :type nvidia_driver: dict
        :rtype: :class:`~dxpy.bindings.dxjob.DXJob`

        Creates a new job that executes the function "main" of this executable with
        the given input *executable_input*.

        '''
        # stage_instance_types, stage_folders, rerun_stages and ignore_reuse_stages are
        # only supported for workflows, but we include them
        # here. Applet-based executables should detect when they
        # receive a truthy workflow-specific value and raise an error.
        run_input = self._get_run_input(executable_input,
                                        project=project,
                                        folder=folder,
                                        name=name,
                                        tags=tags,
                                        properties=properties,
                                        details=details,
                                        instance_type=instance_type,
                                        stage_instance_types=stage_instance_types,
                                        stage_folders=stage_folders,
                                        rerun_stages=rerun_stages,
                                        cluster_spec=cluster_spec,
                                        depends_on=depends_on,
                                        allow_ssh=allow_ssh,
                                        ignore_reuse=ignore_reuse,
                                        ignore_reuse_stages=ignore_reuse_stages,
                                        debug=debug,
                                        delay_workspace_destruction=delay_workspace_destruction,
                                        priority=priority,
                                        head_job_on_demand = head_job_on_demand,
                                        detach=detach,
                                        cost_limit=cost_limit,
                                        rank=rank,
                                        max_tree_spot_wait_time=max_tree_spot_wait_time,
                                        max_job_spot_wait_time=max_job_spot_wait_time,
                                        preserve_job_outputs=preserve_job_outputs,
                                        detailed_job_metrics=detailed_job_metrics,
                                        extra_args=extra_args,
                                        fpga_driver=fpga_driver,
                                        system_requirements=system_requirements,
                                        system_requirements_by_executable=system_requirements_by_executable,
                                        nvidia_driver=nvidia_driver)
        return self._run_impl(run_input, **kwargs)


############
# DXApplet #
############
_applet_required_keys = ['name', 'title', 'summary', 'types', 'tags',
                         'httpsApp', 'properties', 'dxapi', 'inputSpec', 'outputSpec',
                         'runSpec', 'access', 'details']
_applet_optional_keys = ['ignoreReuse', 'treeTurnaroundTimeThreshold']
_applet_describe_output_keys = ['properties', 'details']
_applet_cleanup_keys = ['name', 'title', 'summary', 'types', 'tags',
                        'properties', 'runSpec', 'access', 'details']

def _makeNonexistentAPIWrapper(method):
    def nonexistentAPIWrapper(object_id, input_params=None, always_retry=None, **kwargs):
        raise DXError("Wrapper for " + method + " does not exist")
    return nonexistentAPIWrapper

class DXApplet(DXDataObject, DXExecutable):
    '''
    Remote applet object handler.

    .. py:attribute:: runSpec

       The applet's run specification (a dict indicating, among other things, how the code of the
       applet is to be interpreted). See `the API docs for Run Specification
       <https://documentation.dnanexus.com/developer/api/running-analyses/io-and-run-specifications#run-specification>`_
       for more information.

    .. py:attribute:: dxapi

       String containing the version of the DNAnexus API that the applet should run against.

    .. py:attribute:: access

       The applet's access requirements hash (a dict indicating any nonstandard permissions, such
       as requiring access to the internet, that are needed by the applet). See `the API docs for
       Access Requirements
       <https://documentation.dnanexus.com/developer/api/running-analyses/io-and-run-specifications#access-requirements>`_
       for more information.

    .. py:attribute:: title

       String containing the (human-readable) title of the app

    .. py:attribute:: summary

       String containing a short, one-line summary of the applet's purpose

    .. py:attribute:: description

       String of free-form text (`Markdown <http://daringfireball.net/projects/markdown/>`_ syntax
       is supported) containing a description of the applet. The description is presented to users
       to help them understand the purpose of the app and how to invoke it.

    .. py:attribute:: developerNotes

       String of free-form text (`Markdown <http://daringfireball.net/projects/markdown/>`_ syntax
       is supported) containing information about the internals or implementation details of the
       applet, suitable for developers or advanced users.

    .. automethod:: _new
    '''

    _class = "applet"

    _describe = staticmethod(dxpy.api.applet_describe)
    _add_types = staticmethod(_makeNonexistentAPIWrapper("/applet-xxxx/addTypes"))
    _remove_types = staticmethod(_makeNonexistentAPIWrapper("/applet-xxxx/removeTypes"))
    _get_details = staticmethod(dxpy.api.applet_get_details)
    _set_details = staticmethod(_makeNonexistentAPIWrapper("/applet-xxxx/setDetails"))
    _set_visibility = staticmethod(_makeNonexistentAPIWrapper("/applet-xxxx/setVisibility"))
    _rename = staticmethod(dxpy.api.applet_rename)
    _set_properties = staticmethod(dxpy.api.applet_set_properties)
    _add_tags = staticmethod(dxpy.api.applet_add_tags)
    _remove_tags = staticmethod(dxpy.api.applet_remove_tags)
    _close = staticmethod(_makeNonexistentAPIWrapper("/applet-xxxx/close"))
    _list_projects = staticmethod(dxpy.api.applet_list_projects)

    def _new(self, dx_hash, **kwargs):
        '''
        :param dx_hash: Standard hash populated in :func:`dxpy.bindings.DXDataObject.new()` containing attributes common to all data object classes.
        :type dx_hash: dict
        :param runSpec: Run specification
        :type runSpec: dict
        :param dxapi: API version string
        :type dxapi: string
        :param inputSpec: Input specification (optional)
        :type inputSpec: dict
        :param outputSpec: Output specification (optional)
        :type outputSpec: dict
        :param access: Access specification (optional)
        :type access: dict
        :param title: Title string (optional)
        :type title: string
        :param summary: Summary string (optional)
        :type summary: string
        :param description: Description string (optional)
        :type description: string

        .. note:: It is highly recommended that the higher-level module
           :mod:`dxpy.app_builder` or (preferably) its frontend `dx build
           <https://documentation.dnanexus.com/user/helpstrings-of-sdk-command-line-utilities#build>`_
           be used instead for applet creation.

        Creates an applet with the given parameters. See the API
        documentation for the `/applet/new
        <https://documentation.dnanexus.com/developer/api/running-analyses/applets-and-entry-points#api-method-applet-new>`_
        method for more info. The applet is not run until :meth:`run()`
        is called.

        '''
        for field in 'runSpec', 'dxapi':
            if field not in kwargs:
                raise DXError("%s: Keyword argument %s is required" % (self.__class__.__name__, field))
            dx_hash[field] = kwargs[field]
            del kwargs[field]
        for field in 'inputSpec', 'outputSpec', 'access', 'title', 'summary', 'description':
            if field in kwargs:
                dx_hash[field] = kwargs[field]
                del kwargs[field]

        resp = dxpy.api.applet_new(dx_hash, **kwargs)
        self.set_ids(resp["id"], dx_hash["project"])

    def get(self, **kwargs):
        """
        :returns: Full specification of the remote applet object
        :rtype: dict

        Returns the contents of the applet. The result includes the
        key-value pairs as specified in the API documentation for the
        `/applet-xxxx/get
        <https://documentation.dnanexus.com/developer/api/running-analyses/applets-and-entry-points#api-method-applet-xxxx-get>`_
        method.
        """
        return dxpy.api.applet_get(self._dxid, **kwargs)

    def _run_impl(self, run_input, **kwargs):
        return DXJob(dxpy.api.applet_run(self._dxid, run_input, **kwargs)["id"])

    def _get_run_input(self, executable_input, **kwargs):
        return DXExecutable._get_run_input_fields_for_applet(executable_input, **kwargs)

    def _get_required_keys(self):
        return _applet_required_keys

    def _get_optional_keys(self):
        return _applet_optional_keys

    def _get_describe_output_keys(self):
        return _applet_describe_output_keys

    def _get_cleanup_keys(self):
        return _applet_cleanup_keys

    def run(self, applet_input, *args, **kwargs):
        """
        Creates a new job that executes the function "main" of this applet with
        the given input *applet_input*.

        See :meth:`dxpy.bindings.dxapplet.DXExecutable.run` for the available
        args.
        """
        # Rename applet_input arg to preserve API compatibility when calling
        # DXApplet.run(applet_input=...)
        return super(DXApplet, self).run(applet_input, *args, **kwargs)



from __future__ import annotations
import sys
import csv
import os
import json
from ..exceptions import err_exit


def write_expression_output(
    arg_output,
    arg_delim,
    arg_sql,
    output_listdict_or_string,
    save_uncommon_delim_to_txt=True,
    output_file_name=None,
    error_handler=err_exit,
    colnames=None,
):
    """
    arg_output: str
    A string representing the output file path.
    When it's "-", output is written to stdout.
    This can be directly set as args.output from argparse when calling the function (e.g within dataset_utilities)

    arg_delim: str
    A string representing the delimiter. Defaults to "," when not specified.
    It also determines the file suffix when writing to file.
    This can be set to args.delimiter from argparse when the method is called

    arg_sql: bool
    A boolean representing whether the output_listdict_or_string is a SQL query (string) or not.
    This can be args.sql from argparse

    output_listdict_or_string: 'list of dicts' or 'str' depending on whether arg_sql is False or True, respectively
    This is expected to be the response from vizserver
    if arg_sql is True, this is expected to be a string representing the SQL query
    if arg_sql is False, this is expected to be a list of dicts representing the output of a SQL query
    if output_listdict_or_string is a list of dicts, all dicts must have the same keys which will be used as column names

    save_uncommon_delim_to_txt: bool
    Set this to False if you want to error out when any delimiter other than "," or "\t" is specified

    output_file_name: str
    This is expected to be a record_name which will be used when arg_output is not specified
    Do not append a suffix to this string
    output_file_name is mandatory when arg_output is not specified

    By default delimiter is set "," and file suffix is csv (when writing to file)

    None values are written as empty strings by default (csv.DictWriter behavior)

    """

    IS_OS_WINDOWS = os.name == "nt"
    OS_SPECIFIC_LINE_SEPARATOR = os.linesep
    IS_PYTHON_2 = sys.version_info.major == 2
    IS_PYTHON_3 = sys.version_info.major == 3

    if arg_sql:
        SUFFIX = ".sql"
        if not isinstance(output_listdict_or_string, str):
            error_handler("Expected SQL query to be a string")
    elif arg_delim:
        if arg_delim == ",":
            SUFFIX = ".csv"
        elif arg_delim == "\t":
            SUFFIX = ".tsv"
        else:
            if save_uncommon_delim_to_txt:
                SUFFIX = ".txt"
            else:
                error_handler("Unsupported delimiter: {}".format(arg_delim))
    else:
        SUFFIX = ".csv"

    if arg_output:
        if arg_output == "-":
            WRITE_METHOD = "STDOUT"
        else:
            WRITE_METHOD = "FILE"
            output_file_name = arg_output

    else:
        OUTPUT_DIR = os.getcwd()

        if output_file_name is None:
            error_handler(
                "No output filename specified"
            )  # Developer expected to provide record_name upstream when calling this function

        else:
            WRITE_METHOD = "FILE"
            output_file_name = os.path.join(OUTPUT_DIR, output_file_name + SUFFIX)

    if WRITE_METHOD == "FILE":
        # error out if file already exists or output_file_name is a directory
        if os.path.exists(output_file_name):
            if os.path.isfile(output_file_name):
                error_handler(
                    "{} already exists. Please specify a new file path".format(
                        output_file_name
                    )
                )
            if os.path.isdir(output_file_name):
                error_handler(
                    "{} is a directory. Please specify a new file path".format(
                        output_file_name
                    )
                )

    if arg_sql:
        if WRITE_METHOD == "STDOUT":
            print(output_listdict_or_string)
        elif WRITE_METHOD == "FILE":
            with open(output_file_name, "w") as f:
                f.write(output_listdict_or_string)
        else:
            error_handler("Unexpected error occurred while writing SQL query output")

    else:
        if colnames:
            COLUMN_NAMES = colnames
        else:
            COLUMN_NAMES = output_listdict_or_string[0].keys()

        if not all(
            set(i.keys()) == set(COLUMN_NAMES) for i in output_listdict_or_string
        ):
            error_handler("All rows must have the same column names")

        WRITE_MODE = "wb" if IS_PYTHON_2 or IS_OS_WINDOWS else "w"
        NEWLINE = "" if IS_PYTHON_3 else None
        DELIMITER = str(arg_delim) if arg_delim else ","
        QUOTING = csv.QUOTE_MINIMAL
        QUOTE_CHAR = '"'

        write_args = {
            "mode": WRITE_MODE,
        }

        if IS_PYTHON_3:
            write_args["newline"] = NEWLINE

        dictwriter_params = {
            "fieldnames": COLUMN_NAMES,
            "delimiter": DELIMITER,
            "lineterminator": OS_SPECIFIC_LINE_SEPARATOR,
            "quoting": QUOTING,
            "quotechar": QUOTE_CHAR,
        }

        if WRITE_METHOD == "FILE":
            with open(output_file_name, **write_args) as f:
                w = csv.DictWriter(f, **dictwriter_params)
                w.writeheader()
                w.writerows(output_listdict_or_string)

        elif WRITE_METHOD == "STDOUT":
            w = csv.DictWriter(sys.stdout, **dictwriter_params)
            w.writeheader()
            w.writerows(output_listdict_or_string)

        else:
            error_handler("Unexpected error occurred while writing output")



def pretty_print_json(json_dict: dict) -> str:
    """Pretty-prints the provided JSON object.

    Args:
        json_dict: A string containing valid JSON data.

    Returns:
        Returns a string with formatted JSON or None if there's an error.
    """
    if isinstance(json_dict, dict):
        formatted_json = json.dumps(json_dict, sort_keys=True, indent=4)
        return formatted_json
    else:
        print("WARNING: Invalid JSON provided.", file=sys.stderr)
        return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 DNAnexus, Inc.
#
# This file is part of dx-toolkit (DNAnexus platform client libraries).
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may not
#   use this file except in compliance with the License. You may obtain a copy
#   of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.


# Run manually with python3 src/python/test/test_extract_expression.py TestDXExtractExpression

from __future__ import absolute_import

import unittest
import subprocess
import sys
import os
import dxpy
import copy
import json
import tempfile
import csv
from collections import OrderedDict
import pandas as pd

import shutil
from dxpy_testutil import cd, chdir
from dxpy.bindings.apollo.json_validation_by_schema import JSONValidator
from dxpy.utils.resolver import resolve_existing_path

from dxpy.bindings.apollo.schemas.assay_filtering_json_schemas import (
    EXTRACT_ASSAY_EXPRESSION_JSON_SCHEMA,
)
from dxpy.bindings.apollo.cmd_line_options_validator import ArgsValidator
from dxpy.bindings.apollo.schemas.input_arguments_validation_schemas import (
    EXTRACT_ASSAY_EXPRESSION_INPUT_ARGS_SCHEMA,
)
from dxpy.bindings.apollo.vizclient import VizClient

from dxpy.bindings.apollo.data_transformations import transform_to_expression_matrix
from dxpy.cli.output_handling import write_expression_output
from dxpy.cli.help_messages import EXTRACT_ASSAY_EXPRESSION_JSON_TEMPLATE
from dxpy.bindings.dxrecord import DXRecord
from dxpy.bindings.apollo.dataset import Dataset

from dxpy.bindings.apollo.vizserver_filters_from_json_parser import JSONFiltersValidator
from dxpy.bindings.apollo.schemas.assay_filtering_conditions import (
    EXTRACT_ASSAY_EXPRESSION_FILTERING_CONDITIONS_1_0,
    EXTRACT_ASSAY_EXPRESSION_FILTERING_CONDITIONS_1_1,
    EXTRACT_ASSAY_EXPRESSION_FILTERING_CONDITIONS_1_1_non_optimized,
)
from dxpy.bindings.apollo.vizserver_payload_builder import VizPayloadBuilder
from dxpy.exceptions import err_exit


dirname = os.path.dirname(__file__)

from expression_test_assets.expression_test_input_dict import (
    CLIEXPRESS_TEST_INPUT,
    VIZPAYLOADERBUILDER_TEST_INPUT,
    EXPRESSION_CLI_JSON_FILTERS,
)
from expression_test_assets.expression_test_expected_output_dict import (
    VIZPAYLOADERBUILDER_EXPECTED_OUTPUT,
)


class TestDXExtractExpression(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.maxDiff = None
        test_project_name = "dx-toolkit_test_data"
        cls.proj_id = list(
            dxpy.find_projects(describe=False, level="VIEW", name=test_project_name)
        )[0]["id"]
        cd(cls.proj_id + ":/")
        cls.general_input_dir = os.path.join(dirname, "expression_test_files/input/")
        cls.general_output_dir = os.path.join(dirname, "expression_test_files/output/")
        # Make an output directory if it doesn't already exists
        if not os.path.exists(cls.general_output_dir):
            os.makedirs(cls.general_output_dir)
        cls.json_schema = EXTRACT_ASSAY_EXPRESSION_JSON_SCHEMA
        cls.input_args_schema = EXTRACT_ASSAY_EXPRESSION_INPUT_ARGS_SCHEMA
        cls.cohort_browser_record = (
            cls.proj_id + ":/Extract_Expression/cohort_browser_object"
        )
        cls.expression_dataset_name = "molecular_expression1.dataset"
        cls.expression_dataset = cls.proj_id + ":/" + cls.expression_dataset_name
        cls.combined_expression_cohort_name = "Combined_Expression_Cohort"
        cls.combined_expression_cohort = (
            cls.proj_id + ":/" + cls.combined_expression_cohort_name
        )
        cls.expression_dataset_1_1_name = "molecular_expression_1_1_dataset"
        cls.expression_dataset_1_1 = (
            cls.proj_id + ":/" + cls.expression_dataset_1_1_name
        )
        # In python3, str(type(object)) looks like <{0} 'obj_class'>
        # This impacts our expected error messages
        cls.type_representation = "class"

        cls.default_entity_describe = {
            "id": cls.expression_dataset,
            "project": cls.proj_id,
            "class": "record",
            "sponsored": False,
            "name": "fake_assay",
            "types": ["Dataset"],
            "state": "closed",
            "hidden": False,
            "links": ["database-xxxx", "file-zzzzzz"],
            "folder": "/",
            "tags": [],
            "created": 0,
            "modified": 0,
            "createdBy": {
                "user": "user-test",
                "job": "job-xyz",
                "executable": "app-xyz",
            },
            "size": 0,
            "properties": {},
            "details": {
                "descriptor": {"$dnanexus_link": "file-xyz"},
                "version": "3.0",
                "schema": "ds-molecular_expression_quantification",
                "databases": [{"assay": {"$dnanexus_link": "database-yyyyyy"}}],
                "name": "fake_assay",
                "description": "Dataset: assay",
            },
        }

        cls.default_parser_dict = {
            "apiserver_host": None,
            "apiserver_port": None,
            "apiserver_protocol": None,
            "project_context_id": None,
            "workspace_id": None,
            "security_context": None,
            "auth_token": None,
            "env_help": None,
            "version": None,
            "command": "extract_assay",
            "path": None,
            "list_assays": False,
            "retrieve_expression": False,
            "additional_fields_help": False,
            "assay_name": None,
            "filter_json": None,
            "filter_json_file": None,
            "json_help": False,
            "sql": False,
            "additional_fields": None,
            "expression_matrix": False,
            "delim": None,
            "output": None,
        }

        cls.vizserver_data_mock_response = {
            "results": [
                {
                    "feature_id": "ENST00000450305",
                    "sample_id": "sample_2",
                    "expression": 50,
                    "strand": "+",
                },
                {
                    "feature_id": "ENST00000456328",
                    "sample_id": "sample_2",
                    "expression": 90,
                    "strand": "+",
                },
                {
                    "feature_id": "ENST00000488147",
                    "sample_id": "sample_2",
                    "expression": 20,
                    "strand": "-",
                },
            ]
        }
        cls.argparse_expression_help_message = os.path.join(
            dirname, "help_messages/extract_expression_help_message.txt"
        )

    @classmethod
    def input_arg_error_handler(cls, message):
        raise ValueError(message)

    @classmethod
    def json_error_handler(cls, message):
        raise ValueError(message)

    @classmethod
    def common_value_error_handler(cls, message):
        raise ValueError(message)

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.general_output_dir)

    #
    # Helper functions used by different types of tests
    #

    def common_negative_filter_test(self, json_name, expected_error_message):
        input_json = CLIEXPRESS_TEST_INPUT["malformed"][json_name]
        validator = JSONValidator(
            self.json_schema, error_handler=self.json_error_handler
        )

        with self.assertRaises(ValueError) as cm:
            validator.validate(input_json)

        self.assertEqual(expected_error_message, str(cm.exception).strip())

    def common_positive_filter_test(self, json_name):
        input_json = CLIEXPRESS_TEST_INPUT["valid"][json_name]

        validator = JSONValidator(
            self.json_schema, error_handler=self.json_error_handler
        )

        validator.validate(input_json)

    def common_input_args_test(self, input_argument_dict, expected_error_message):
        # Deep copy the default parser dictionary
        parser_dict = {key: value for key, value in self.default_parser_dict.items()}
        for input_argument in input_argument_dict:
            if input_argument in self.default_parser_dict:
                parser_dict[input_argument] = input_argument_dict[input_argument]
            else:
                print("unrecognized argument in input args")
                return False

        input_arg_validator = ArgsValidator(
            parser_dict,
            EXTRACT_ASSAY_EXPRESSION_INPUT_ARGS_SCHEMA,
            error_handler=self.input_arg_error_handler,
        )
        with self.assertRaises(ValueError) as cm:
            input_arg_validator.validate_input_combination()

        self.assertEqual(expected_error_message, str(cm.exception).strip())

    #
    # Expression matrix tests
    #

    def test_basic_exp_matrix_transform(self):
        vizserver_results = [
            {
                "feature_id": "ENST00000450305",
                "sample_id": "sample_2",
                "expression": 50,
            },
            {
                "feature_id": "ENST00000456328",
                "sample_id": "sample_2",
                "expression": 90,
            },
            {
                "feature_id": "ENST00000488147",
                "sample_id": "sample_2",
                "expression": 20,
            },
        ]
        expected_output = [
            {
                "ENST00000450305": 50,
                "ENST00000456328": 90,
                "ENST00000488147": 20,
                "sample_id": "sample_2",
            }
        ]

        transformed_results, colnames = transform_to_expression_matrix(
            vizserver_results
        )
        self.assertEqual(expected_output, transformed_results)

    def test_two_sample_exp_transform(self):
        vizserver_results = [
            {
                "feature_id": "ENST00000450305",
                "sample_id": "sample_2",
                "expression": 50,
            },
            {
                "feature_id": "ENST00000456328",
                "sample_id": "sample_1",
                "expression": 90,
            },
            {
                "feature_id": "ENST00000488147",
                "sample_id": "sample_2",
                "expression": 20,
            },
        ]

        expected_output = [
            {
                "sample_id": "sample_2",
                "ENST00000450305": 50,
                "ENST00000488147": 20,
                "ENST00000456328": None,
            },
            {
                "sample_id": "sample_1",
                "ENST00000456328": 90,
                "ENST00000450305": None,
                "ENST00000488147": None,
            },
        ]

        transformed_results, colnames = transform_to_expression_matrix(
            vizserver_results
        )
        self.assertEqual(expected_output, transformed_results)

    def test_two_sample_feat_id_overlap_exp_trans(self):
        vizserver_results = [
            {
                "feature_id": "ENST00000450305",
                "sample_id": "sample_2",
                "expression": 50,
            },
            {
                "feature_id": "ENST00000450305",
                "sample_id": "sample_1",
                "expression": 77,
            },
            {
                "feature_id": "ENST00000456328",
                "sample_id": "sample_1",
                "expression": 90,
            },
            {
                "feature_id": "ENST00000488147",
                "sample_id": "sample_2",
                "expression": 20,
            },
        ]
        expected_output = [
            {
                "sample_id": "sample_2",
                "ENST00000450305": 50,
                "ENST00000488147": 20,
                "ENST00000456328": None,
            },
            {
                "sample_id": "sample_1",
                "ENST00000450305": 77,
                "ENST00000456328": 90,
                "ENST00000488147": None,
            },
        ]

        transformed_results, colnames = transform_to_expression_matrix(
            vizserver_results
        )
        self.assertEqual(expected_output, transformed_results)

    def test_exp_transform_output_compatibility(self):
        vizserver_results = [
            {
                "feature_id": "ENST00000450305",
                "sample_id": "sample_2",
                "expression": 50,
            },
            {
                "feature_id": "ENST00000450305",
                "sample_id": "sample_1",
                "expression": 77,
            },
            {
                "feature_id": "ENST00000456328",
                "sample_id": "sample_1",
                "expression": 90,
            },
            {
                "feature_id": "ENST00000488147",
                "sample_id": "sample_2",
                "expression": 20,
            },
        ]

        # The replace statement removes tabs(actually blocks of 4 spaces) that have been inserted
        # for readability in this python file
        expected_result = """sample_id,ENST00000450305,ENST00000456328,ENST00000488147
                             sample_2,50,,20
                             sample_1,77,90,""".replace(
            " ", ""
        )

        transformed_results, colnames = transform_to_expression_matrix(
            vizserver_results
        )
        output_path = os.path.join(self.general_output_dir, "exp_transform_compat.csv")
        # Generate the formatted output file
        write_expression_output(
            output_path, ",", False, transformed_results, colnames=colnames
        )

        with open(output_path, "r") as infile:
            data = infile.read()
        self.assertEqual(expected_result.strip(), data.strip())

    #
    # Positive output tests
    #

    def test_output_data_format(self):
        expected_result = """feature_id,sample_id,expression,strand
            ENST00000450305,sample_2,50,+
            ENST00000456328,sample_2,90,+
            ENST00000488147,sample_2,20,-""".replace(
            " ", ""
        )
        output_path = os.path.join(
            self.general_output_dir, "extract_assay_expression_data.csv"
        )
        # Generate the formatted output file
        write_expression_output(
            output_path,
            ",",
            False,
            self.vizserver_data_mock_response["results"],
        )
        # Read the output file back in and compare to expected result
        # Since the test should fail if the formatting is wrong, not just if the data is wrong, we
        # can do a simple string comparison
        with open(output_path, "r") as infile:
            data = infile.read()
        self.assertEqual(expected_result.strip(), data.strip())

    def test_output_sql_format(self):
        sql_mock_response = {
            "sql": "SELECT `expression_1`.`feature_id` AS `feature_id`, `expression_1`.`sample_id` AS `sample_id`, `expression_1`.`value` AS `expression`, `expr_annotation_1`.`strand` AS `strand` FROM `database_gypg8qq06j8kzzp2yybfbzfk__enst_short_multiple_assays2`.`expression` AS `expression_1` LEFT OUTER JOIN `database_gypg8qq06j8kzzp2yybfbzfk__enst_short_multiple_assays2`.`expr_annotation` AS `expr_annotation_1` ON `expression_1`.`feature_id` = `expr_annotation_1`.`feature_id` WHERE `expression_1`.`value` >= 1 AND `expr_annotation_1`.`chr` = '1' AND (`expr_annotation_1`.`end` BETWEEN 7 AND 250000000 OR `expr_annotation_1`.`start` BETWEEN 7 AND 250000000 OR `expr_annotation_1`.`end` >= 250000000 AND `expr_annotation_1`.`start` <= 7)"
        }
        expected_result = "SELECT `expression_1`.`feature_id` AS `feature_id`, `expression_1`.`sample_id` AS `sample_id`, `expression_1`.`value` AS `expression`, `expr_annotation_1`.`strand` AS `strand` FROM `database_gypg8qq06j8kzzp2yybfbzfk__enst_short_multiple_assays2`.`expression` AS `expression_1` LEFT OUTER JOIN `database_gypg8qq06j8kzzp2yybfbzfk__enst_short_multiple_assays2`.`expr_annotation` AS `expr_annotation_1` ON `expression_1`.`feature_id` = `expr_annotation_1`.`feature_id` WHERE `expression_1`.`value` >= 1 AND `expr_annotation_1`.`chr` = '1' AND (`expr_annotation_1`.`end` BETWEEN 7 AND 250000000 OR `expr_annotation_1`.`start` BETWEEN 7 AND 250000000 OR `expr_annotation_1`.`end` >= 250000000 AND `expr_annotation_1`.`start` <= 7)"
        output_path = os.path.join(
            self.general_output_dir, "extract_assay_expression_sql.csv"
        )
        # Generate the formatted output file
        write_expression_output(
            arg_output=output_path,
            arg_delim=",",
            arg_sql=True,
            output_listdict_or_string=sql_mock_response["sql"],
        )
        # Read the output file back in and compare to expected result
        # Since the test should fail if the formatting is wrong, not just if the data is wrong, we
        # can do a simple string comparison
        with open(output_path, "r") as infile:
            data = infile.read()
        self.assertEqual(expected_result.strip(), data.strip())

    #
    # Negative output tests
    #

    def test_output_sql_not_string(self):
        expected_error_message = "Expected SQL query to be a string"
        with self.assertRaises(ValueError) as cm:
            write_expression_output(
                arg_output="-",
                arg_delim=",",
                arg_sql=True,
                output_listdict_or_string=["not a string-formatted SQL query"],
                error_handler=self.common_value_error_handler,
            )
        err_msg = str(cm.exception).strip()
        self.assertEqual(expected_error_message, err_msg)

    def test_output_bad_delimiter(self):
        bad_delim = "|"
        expected_error_message = "Unsupported delimiter: {}".format(bad_delim)
        with self.assertRaises(ValueError) as cm:
            write_expression_output(
                arg_output="-",
                arg_delim=bad_delim,
                arg_sql=False,
                output_listdict_or_string=self.vizserver_data_mock_response["results"],
                save_uncommon_delim_to_txt=False,
                error_handler=self.common_value_error_handler,
            )
        err_msg = str(cm.exception).strip()
        self.assertEqual(expected_error_message, err_msg)

    # EM-14
    def test_output_already_exist(self):
        output_path = os.path.join(
            self.general_output_dir, "already_existing_output.csv"
        )
        expected_error_message = (
            "{} already exists. Please specify a new file path".format(output_path)
        )

        with open(output_path, "w") as outfile:
            outfile.write("this output file already created")

        with self.assertRaises(ValueError) as cm:
            write_expression_output(
                arg_output=output_path,
                arg_delim=",",
                arg_sql=False,
                output_listdict_or_string=self.vizserver_data_mock_response["results"],
                save_uncommon_delim_to_txt=False,
                error_handler=self.common_value_error_handler,
            )

        err_msg = str(cm.exception).strip()
        self.assertEqual(expected_error_message, err_msg)

    def test_output_is_directory(self):
        output_path = os.path.join(self.general_output_dir, "directory")
        expected_error_message = (
            "{} is a directory. Please specify a new file path".format(output_path)
        )
        os.mkdir(output_path)
        with self.assertRaises(ValueError) as cm:
            write_expression_output(
                arg_output=output_path,
                arg_delim=",",
                arg_sql=False,
                output_listdict_or_string=self.vizserver_data_mock_response["results"],
                save_uncommon_delim_to_txt=False,
                error_handler=self.common_value_error_handler,
            )

        err_msg = str(cm.exception).strip()
        self.assertEqual(expected_error_message, err_msg)

    # EM-1
    # Test PATH argument not provided
    def test_path_missing(self):
        input_dict = {}
        expected_error_message = (
            'At least one of the following arguments is required: "Path", "--json-help"'
        )
        self.common_input_args_test(input_dict, expected_error_message)

    # EM-10
    # When --additional-fields-help is presented with other options
    def test_additional_fields_help_other_options(self):
        expected_error_message = '"--additional-fields-help" cannot be passed with any option other than "--retrieve-expression".'
        input_dict = {
            "path": self.expression_dataset,
            "assay_name": "test_assay",
            "additional_fields_help": True,
        }
        self.common_input_args_test(input_dict, expected_error_message)

    # EM-11
    # When invalid additional fields are passed
    def invalid_additional_fields(self):
        expected_error_message = "One or more of the supplied fields using --additional-fields are invalid. Please run --additional-fields-help for a list of valid fields"
        input_dict = {
            "path": self.expression_dataset,
            "retrieve_expression": True,
            "filter_json": r'{"annotation": {"feature_id": ["ENSG0000001", "ENSG00000002"]}}',
            "additional_fields": "feature_name,bad_field",
        }
        self.common_input_args_test(input_dict, expected_error_message)

    # EM-12
    # When –list-assays is presented with other options
    def test_list_assays_assay_name(self):
        expected_error_message = (
            '"--list-assays" cannot be presented with other options'
        )
        input_dict = {
            "path": self.expression_dataset,
            "list_assays": True,
            "assay_name": "fake_assay",
        }
        self.common_input_args_test(input_dict, expected_error_message)

    # EM-17
    # When the .json file provided does not exist
    def test_json_file_not_exist(self):
        missing_json_path = os.path.join(self.general_input_dir, "nonexistent.json")
        expected_error_message = (
            "JSON file {} provided to --retrieve-expression does not exist".format(
                missing_json_path
            )
        )
        command = [
            "dx",
            "extract_assay",
            "expression",
            self.expression_dataset,
            "--retrieve-expression",
            "--filter-json-file",
            missing_json_path,
        ]

        process = subprocess.Popen(
            command, stderr=subprocess.PIPE, universal_newlines=True
        )
        actual_err_msg = process.communicate()[1]
        # print(actual_err_msg)

        self.assertIn(expected_error_message, actual_err_msg)

    # EM-21
    # When --json-help is passed with another option from --assay-name, --sql, --additional-fields, --expression-matix, --output
    def test_json_help_other_option(self):
        expected_error_message = '"--json-help" cannot be passed with any option other than "--retrieve-expression".'
        input_dict = {
            "path": self.expression_dataset,
            "json_help": True,
            "assay_name": "test_assay",
        }
        self.common_input_args_test(input_dict, expected_error_message)

    # EM-23
    # --expression-matrix/-em cannot be used with --sql
    def test_exp_matrix_sql(self):
        expected_error_message = (
            '"--expression-matrix"/"-em" cannot be passed with the flag, "--sql".'
        )
        input_dict = {
            "path": self.expression_dataset,
            "expression_matrix": True,
            "retrieve_expression": True,
            "filter_json": r'{"annotation": {"feature_name": ["BRCA2"]}}',
            "sql": True,
        }
        self.common_input_args_test(input_dict, expected_error_message)

    # Malformed input json tests
    # EM-18, EM-19, EM-20
    #

    def test_annotation_conflicting_keys(self):
        self.common_negative_filter_test(
            "annotation_conflicting_keys",
            "For annotation, exactly one of feature_name or feature_id must be provided in the supplied JSON object.",
        )

    def test_annotation_id_type(self):
        self.common_negative_filter_test(
            "annotation_id_type",
            "Key 'feature_id' has an invalid type. Expected <{0} 'list'> but got <{0} 'dict'>".format(
                self.type_representation
            ).format(
                self.type_representation
            ),
        )

    def test_annotation_name_type(self):
        self.common_negative_filter_test(
            "annotation_name_type",
            "Key 'feature_name' has an invalid type. Expected <{0} 'list'> but got <{0} 'dict'>".format(
                self.type_representation
            ),
        )

    def test_annotation_type(self):
        self.common_negative_filter_test(
            "annotation_type",
            "Key 'annotation' has an invalid type. Expected <{0} 'dict'> but got <{0} 'list'>".format(
                self.type_representation
            ),
        )

    def test_bad_dependent_conditional(self):
        self.common_negative_filter_test(
            "bad_dependent_conditional",
            "When expression is present, one of the following keys must be also present: annotation, location.",
        )

    def test_bad_toplevel_key(self):
        self.common_negative_filter_test(
            "bad_toplevel_key", "Found following invalid filters: ['not_real_key']"
        )

    def test_conflicting_toplevel(self):
        self.common_negative_filter_test(
            "conflicting_toplevel",
            "Exactly one of location or annotation must be provided in the supplied JSON object.",
        )

    # EM-15
    def test_empty_dict(self):
        self.common_negative_filter_test(
            "empty_dict", "Input JSON must be a non-empty dict."
        )

    def test_expression_max_type(self):
        self.common_negative_filter_test(
            "expression_max_type",
            "Key 'max_value' has an invalid type. Expected (<{0} 'int'>, <{0} 'float'>) but got <{0} 'str'>".format(
                self.type_representation
            ),
        )

    def test_expression_min_type(self):
        self.common_negative_filter_test(
            "expression_min_type",
            "Key 'min_value' has an invalid type. Expected (<{0} 'int'>, <{0} 'float'>) but got <{0} 'str'>".format(
                self.type_representation
            ),
        )

    def test_expression_type(self):
        self.common_negative_filter_test(
            "expression_type",
            "Key 'expression' has an invalid type. Expected <{0} 'dict'> but got <{0} 'list'>".format(
                self.type_representation
            ),
        )

    def test_location_chrom_type(self):
        self.common_negative_filter_test(
            "location_chrom_type",
            "Key 'chromosome' has an invalid type. Expected <{0} 'str'> but got <{0} 'int'>".format(
                self.type_representation
            ),
        )

    def test_location_end_type(self):
        self.common_negative_filter_test(
            "location_end_type",
            "Key 'ending_position' has an invalid type. Expected <{0} 'str'> but got <{0} 'int'>".format(
                self.type_representation
            ),
        )

    def test_location_item_type(self):
        self.common_negative_filter_test(
            "location_item_type",
            "Expected list items within 'location' to be of type <{0} 'dict'> but got <{0} 'list'> instead.".format(
                self.type_representation
            ),
        )

    def test_location_missing_chr(self):
        self.common_negative_filter_test(
            "location_missing_chr",
            "Required key 'chromosome' was not found in the input JSON.",
        )

    def test_location_missing_end(self):
        self.common_negative_filter_test(
            "location_missing_end",
            "Required key 'ending_position' was not found in the input JSON.",
        )

    def test_location_missing_start(self):
        self.common_negative_filter_test(
            "location_missing_start",
            "Required key 'starting_position' was not found in the input JSON.",
        )

    def test_location_start_type(self):
        self.common_negative_filter_test(
            "location_start_type",
            "Key 'starting_position' has an invalid type. Expected <{0} 'str'> but got <{0} 'int'>".format(
                self.type_representation
            ),
        )

    def test_location_type(self):
        self.common_negative_filter_test(
            "location_type",
            "Key 'location' has an invalid type. Expected <{0} 'list'> but got <{0} 'dict'>".format(
                self.type_representation
            ),
        )

    def test_sample_id_type(self):
        self.common_negative_filter_test(
            "sample_id_type",
            "Key 'sample_id' has an invalid type. Expected <{0} 'list'> but got <{0} 'dict'>".format(
                self.type_representation
            ),
        )

    # EM-20
    def test_sample_id_only(self):
        self.common_negative_filter_test(
            "sample_id_only",
            "When sample_id is present, one of the following keys must be also present: expression, annotation, location.",
        )

    def test_sample_id_location_annotation(self):
        self.common_negative_filter_test(
            "sample_id_location_annotation",
            "Exactly one of location or annotation must be provided in the supplied JSON object.",
        )

    #
    # Correct JSON inputs
    #

    def test_annotation_feature_id(self):
        self.common_positive_filter_test("annotation_feature_id")

    def test_annotation_feature_name(self):
        self.common_positive_filter_test("annotation_feature_name")

    def test_dependent_conditional_annotation(self):
        self.common_positive_filter_test("dependent_conditional_annotation")

    def test_dependent_conditional_location(self):
        self.common_positive_filter_test("dependent_conditional_location")

    def test_expression_max_only(self):
        self.common_positive_filter_test("expression_max_only")

    def test_expression_min_and_max(self):
        self.common_positive_filter_test("expression_min_and_max")

    def test_expression_min_only(self):
        self.common_positive_filter_test("expression_min_only")

    def test_multi_location(self):
        self.common_positive_filter_test("multi_location")

    def test_single_location(self):
        self.common_positive_filter_test("single_location")

    ##### Test argparse's --help output
    def test_argparse_help_txt(self):
        expected_result = self.argparse_expression_help_message
        with open(expected_result) as f:
            # lines = f.readlines()
            file = f.read()
        process = subprocess.check_output("dx extract_assay expression -h", shell=True)
        help_output = process.decode()

        # self.assertEqual(file,help_output) fails due to some differences in where linebreaks appear in the text
        self.assertEqual(
            file.replace(" ", "").replace("\n", ""),
            help_output.replace(" ", "").replace("\n", ""),
        )

    #### Test --json-help
    def test_json_help_template(self):
        process = subprocess.check_output(
            "dx extract_assay expression --retrieve-expression fakepath --json-help",
            shell=True,
        )
        self.assertIn(EXTRACT_ASSAY_EXPRESSION_JSON_TEMPLATE, process.decode())
        self.assertIn(
            "Additional descriptions of filtering keys and permissible values",
            process.decode(),
        )

    def load_record_via_dataset_class(self, record_path):
        _, _, entity = resolve_existing_path(record_path)
        entity_describe = entity["describe"]
        record = DXRecord(entity_describe["id"], entity_describe["project"])
        dataset, cohort_info = Dataset.resolve_cohort_to_dataset(record)

        return dataset, cohort_info, record

    def test_dataset_class_basic(self):
        dataset, cohort, record = self.load_record_via_dataset_class(
            self.expression_dataset
        )

        record_details = record.describe(
            default_fields=True, fields={"properties", "details"}
        )

        self.assertIsNone(cohort)
        self.assertEqual(
            dataset.descriptor_file_dict["name"], self.expression_dataset_name
        )
        self.assertIn("vizserver", dataset.visualize_info["url"])
        self.assertEqual("3.0", dataset.visualize_info["version"])
        self.assertEqual("3.0", dataset.visualize_info["datasetVersion"])
        self.assertEqual(
            dataset.descriptor_file,
            record_details["details"]["descriptor"]["$dnanexus_link"],
        )
        self.assertIn(
            "molecular_expression1", dataset.assay_names_list("molecular_expression")
        )
        self.assertEqual(dataset.detail_describe["types"], record_details["types"])

    def test_dataset_class_basic_1_1(self):
        dataset, cohort, record = self.load_record_via_dataset_class(
            self.expression_dataset_1_1
        )

        record_details = record.describe(
            default_fields=True, fields={"properties", "details"}
        )

        self.assertIsNone(cohort)
        self.assertIn("vizserver", dataset.visualize_info["url"])
        self.assertEqual("3.0", dataset.visualize_info["version"])
        self.assertEqual("3.0", dataset.visualize_info["datasetVersion"])
        self.assertEqual(
            dataset.descriptor_file,
            record_details["details"]["descriptor"]["$dnanexus_link"],
        )
        self.assertIn(
            "molecular_expression_1_1", dataset.assay_names_list("molecular_expression")
        )
        assert (
            dataset.descriptor_file_dict["assays"][0]["generalized_assay_model_version"]
            == "1.1"
        )

    def test_dataset_class_cohort_resolution(self):
        dataset, cohort, record = self.load_record_via_dataset_class(
            self.combined_expression_cohort
        )

        record_details = record.describe(
            default_fields=True, fields={"properties", "details"}
        )
        expected_dataset_id = record_details["details"]["dataset"]["$dnanexus_link"]
        expected_dataset_describe = DXRecord(expected_dataset_id).describe(
            default_fields=True, fields={"properties", "details"}
        )
        expected_descriptor_id = expected_dataset_describe["details"]["descriptor"][
            "$dnanexus_link"
        ]

        self.assertIsNotNone(cohort)
        self.assertIn("SELECT `sample_id`", cohort["details"]["baseSql"])
        self.assertIn("pheno_filters", cohort["details"]["filters"])
        self.assertIn("CohortBrowser", cohort["types"])
        self.assertEqual(dataset.get_id(), expected_dataset_id)
        self.assertEqual(dataset.descriptor_file, expected_descriptor_id)
        self.assertIn(
            "molecular_expression1", dataset.assay_names_list("molecular_expression")
        )
        self.assertEqual(
            "molecular_expression",
            dataset.descriptor_file_dict["assays"][0]["generalized_assay_model"],
        )
        self.assertIn("Dataset", dataset.detail_describe["types"])
        self.assertIn("vizserver", dataset.vizserver_url)

    ### Test VizPayloadBuilder Class
    ### tests data model 1.0 and 1.1
    ### For 1.1: when location EXTRACT_ASSAY_EXPRESSION_FILTERING_CONDITIONS_1_1_non_optimized is used
    ### otherwise EXTRACT_ASSAY_EXPRESSION_FILTERING_CONDITIONS_1_1

    # Genomic location filters
    # genomic + cohort

    # 1.0
    def test_vizpayloadbuilder_location_cohort(self):
        self.common_vizpayloadbuilder_test_helper_method(
            self.combined_expression_cohort, "test_vizpayloadbuilder_location_cohort"
        )

    def test_vizpayloadbuilder_location_multiple(self):
        self.common_vizpayloadbuilder_test_helper_method(
            self.expression_dataset, "test_vizpayloadbuilder_location_multiple"
        )

    # 1.1
    def test_vizpayloadbuilder_location_1_1(self):
        self.common_vizpayloadbuilder_test_helper_method(
            record_path=self.expression_dataset_1_1,
            test_name="test_vizpayloadbuilder_location_cohort",
            schema=EXTRACT_ASSAY_EXPRESSION_FILTERING_CONDITIONS_1_1_non_optimized,
            model_version=1.1,
        )

    # Annotation filters
    def test_vizpayloadbuilder_annotation_feature_name(self):
        self.common_vizpayloadbuilder_test_helper_method(
            self.expression_dataset, "test_vizpayloadbuilder_annotation_feature_name"
        )

    def test_vizpayloadbuilder_annotation_feature_id(self):
        self.common_vizpayloadbuilder_test_helper_method(
            self.expression_dataset, "test_vizpayloadbuilder_annotation_feature_id"
        )

    # 1.1
    def test_vizpayloadbuilder_annotation_feature_name_1_1(self):
        self.common_vizpayloadbuilder_test_helper_method(
            record_path=self.expression_dataset_1_1,
            test_name="test_vizpayloadbuilder_annotation_feature_name",
            schema=EXTRACT_ASSAY_EXPRESSION_FILTERING_CONDITIONS_1_1,
            model_version=1.1,
        )

    # Expression filters (with location or annotation)
    # expression + annotation - ID
    def test_vizpayloadbuilder_expression_min(self):
        self.common_vizpayloadbuilder_test_helper_method(
            self.expression_dataset, "test_vizpayloadbuilder_expression_min"
        )

    # expression + annotation - name
    def test_vizpayloadbuilder_expression_max(self):
        self.common_vizpayloadbuilder_test_helper_method(
            self.expression_dataset, "test_vizpayloadbuilder_expression_max"
        )

    # expression + location
    def test_vizpayloadbuilder_expression_mixed(self):
        self.common_vizpayloadbuilder_test_helper_method(
            self.expression_dataset, "test_vizpayloadbuilder_expression_mixed"
        )

    # Sample filter
    def test_vizpayloadbuilder_sample(self):
        self.common_vizpayloadbuilder_test_helper_method(
            self.expression_dataset, "test_vizpayloadbuilder_sample", data_test=False
        )

    # Annotation_expression_sample_id (using flattened filters)
    def test_vizpayloadbuilder_annotation_expression_sample_id(self):
        self.common_vizpayloadbuilder_test_helper_method(
            record_path=self.expression_dataset,
            test_name="test_vizpayloadbuilder_sample_annotation_expression_sample_id",
        )

    # 1.1
    def test_vizpayloadbuilder_annotation_expression_sample_id_1_1(self):
        self.common_vizpayloadbuilder_test_helper_method(
            record_path=self.expression_dataset_1_1,
            test_name="test_vizpayloadbuilder_sample_annotation_expression_sample_id",
            schema=EXTRACT_ASSAY_EXPRESSION_FILTERING_CONDITIONS_1_1,
            model_version=1.1,
        )

    # General (mixed) filters
    def test_vizpayloadbuilder_location_sample_expression(self):
        self.common_vizpayloadbuilder_test_helper_method(
            self.expression_dataset,
            "test_vizpayloadbuilder_location_sample_expression",
            data_test=False,
        )

    def test_vizpayloadbuilder_annotation_sample_expression(self):
        self.common_vizpayloadbuilder_test_helper_method(
            self.expression_dataset,
            "test_vizpayloadbuilder_annotation_sample_expression",
            data_test=False,
        )

    def common_vizpayloadbuilder_test_helper_method(
        self,
        record_path,
        test_name,
        data_test=True,
        schema=EXTRACT_ASSAY_EXPRESSION_FILTERING_CONDITIONS_1_0,
        model_version=1.0,
    ):
        _, _, entity = resolve_existing_path(record_path)
        entity_describe = entity["describe"]
        record_id = entity_describe["id"]

        record = DXRecord(record_id)
        dataset, cohort_info = Dataset.resolve_cohort_to_dataset(record)
        dataset_id = dataset.dataset_id

        if cohort_info:
            BASE_SQL = cohort_info.get("details").get("baseSql")
            COHORT_FILTERS = cohort_info.get("details").get("filters")
            IS_COHORT = True
        else:
            BASE_SQL = None
            COHORT_FILTERS = None
            IS_COHORT = False

        url = dataset.vizserver_url
        project = dataset.project_id

        # vizserver_filters_from_json_parser.JSONFiltersValidator using the CLIEXPRESS schema
        _db_columns_list = schema["output_fields_mapping"].get("default")

        # JSONFiltersValidator to build the complete payload
        json_input = VIZPAYLOADERBUILDER_TEST_INPUT[test_name]
        input_json_parser = JSONFiltersValidator(json_input, schema)
        vizserver_raw_filters = input_json_parser.parse()

        # VizClient to submit the payload and get a response
        client = VizClient(url, project)

        viz = VizPayloadBuilder(
            project_context=project,
            output_fields_mapping=_db_columns_list,
            filters={"filters": COHORT_FILTERS} if IS_COHORT else None,
            order_by=schema["order_by"],
            limit=None,
            base_sql=BASE_SQL,
            is_cohort=IS_COHORT,
            error_handler=err_exit,
        )

        assay_1_name = dataset.descriptor_file_dict["assays"][0]["name"]
        assay_1_id = dataset.descriptor_file_dict["assays"][0]["uuid"]

        viz.assemble_assay_raw_filters(
            assay_name=assay_1_name, assay_id=assay_1_id, filters=vizserver_raw_filters
        )
        vizserver_payload = viz.build()

        vizserver_response_data = client.get_data(vizserver_payload, dataset_id)[
            "results"
        ]
        vizserver_response_sql = client.get_raw_sql(vizserver_payload, dataset_id)[
            "sql"
        ]

        data_output = vizserver_response_data
        sql_output = vizserver_response_sql

        if data_test:
            exp_data_output = VIZPAYLOADERBUILDER_EXPECTED_OUTPUT[test_name][
                "expected_data_output"
            ]
            # assertCountEqual asserts that two iterables have the same elements, ignoring order
            self.assertCountEqual(data_output, exp_data_output)

        if model_version == 1.0:
            expected_sql = "expected_sql_output"
        if model_version == 1.1:
            expected_sql = "expected_sql_output_1_1"

        exp_sql_output = VIZPAYLOADERBUILDER_EXPECTED_OUTPUT[test_name][expected_sql]
        if isinstance(exp_sql_output, list):
            # Some of the sub-queries may have slightly different order due to the way the keys are ordered in the payload dict
            # In other words, the queries are still correct, but the order of sub-queries may be different
            self.assertIn(sql_output, exp_sql_output)
        else:
            self.assertEqual(sql_output, exp_sql_output)

    def run_dx_extract_assay_expression_cmd(
        self,
        dataset_or_cohort,
        filters_json,
        additional_fields,
        sql,
        output="-",
        extra_args=None,
        subprocess_run=False,
    ):
        command = [
            "dx",
            "extract_assay",
            "expression",
            dataset_or_cohort,
            "--retrieve-expression",
            "--filter-json",
            str(filters_json).replace("'", '"'),
            "-o",
            output,
        ]

        if sql:
            command.append("--sql")

        if additional_fields:
            command.extend(["--additional-fields", additional_fields])

        if extra_args:
            command.append(extra_args)

        if subprocess_run:
            process = subprocess.run(
                command, capture_output=True, text=True, check=False
            )

        else:
            process = subprocess.check_output(
                command,
                universal_newlines=True,
            )

        return process

    def test_dx_extract_cmd_location_expression_sample_sql(self):
        expected_sql_query = [
            "SELECT `expression_1`.`feature_id` AS `feature_id`, `expression_1`.`sample_id` AS `sample_id`, `expression_1`.`value` AS `expression`, `expr_annotation_1`.`gene_name` AS `feature_name`, `expr_annotation_1`.`chr` AS `chrom`, `expr_annotation_1`.`start` AS `start` FROM `database_gzky7400vgpyzy621q43gkkf__molecular_expression1_db`.`expression` AS `expression_1` LEFT OUTER JOIN `database_gzky7400vgpyzy621q43gkkf__molecular_expression1_db`.`expr_annotation` AS `expr_annotation_1` ON `expression_1`.`feature_id` = `expr_annotation_1`.`feature_id` WHERE (`expr_annotation_1`.`chr` = '11' AND (`expr_annotation_1`.`start` BETWEEN 8693350 AND 67440200 OR `expr_annotation_1`.`end` BETWEEN 8693350 AND 67440200 OR `expr_annotation_1`.`start` <= 8693350 AND `expr_annotation_1`.`end` >= 67440200) OR `expr_annotation_1`.`chr` = 'X' AND (`expr_annotation_1`.`start` BETWEEN 148500700 AND 148994424 OR `expr_annotation_1`.`end` BETWEEN 148500700 AND 148994424 OR `expr_annotation_1`.`start` <= 148500700 AND `expr_annotation_1`.`end` >= 148994424) OR `expr_annotation_1`.`chr` = '17' AND (`expr_annotation_1`.`start` BETWEEN 75228160 AND 75235759 OR `expr_annotation_1`.`end` BETWEEN 75228160 AND 75235759 OR `expr_annotation_1`.`start` <= 75228160 AND `expr_annotation_1`.`end` >= 75235759)) AND `expression_1`.`value` >= 25.63 AND `expression_1`.`sample_id` IN ('sample_1', 'sample_2') ORDER BY `feature_id` ASC, `sample_id` ASC",
            "SELECT `expression_1`.`feature_id` AS `feature_id`, `expression_1`.`sample_id` AS `sample_id`, `expression_1`.`value` AS `expression`, `expr_annotation_1`.`gene_name` AS `feature_name`, `expr_annotation_1`.`chr` AS `chrom`, `expr_annotation_1`.`start` AS `start` FROM `database_gzky7400vgpyzy621q43gkkf__molecular_expression1_db`.`expression` AS `expression_1` LEFT OUTER JOIN `database_gzky7400vgpyzy621q43gkkf__molecular_expression1_db`.`expr_annotation` AS `expr_annotation_1` ON `expression_1`.`feature_id` = `expr_annotation_1`.`feature_id` WHERE (`expr_annotation_1`.`chr` = '11' AND (`expr_annotation_1`.`end` BETWEEN 8693350 AND 67440200 OR `expr_annotation_1`.`start` BETWEEN 8693350 AND 67440200 OR `expr_annotation_1`.`end` >= 67440200 AND `expr_annotation_1`.`start` <= 8693350) OR `expr_annotation_1`.`chr` = 'X' AND (`expr_annotation_1`.`end` BETWEEN 148500700 AND 148994424 OR `expr_annotation_1`.`start` BETWEEN 148500700 AND 148994424 OR `expr_annotation_1`.`end` >= 148994424 AND `expr_annotation_1`.`start` <= 148500700) OR `expr_annotation_1`.`chr` = '17' AND (`expr_annotation_1`.`end` BETWEEN 75228160 AND 75235759 OR `expr_annotation_1`.`start` BETWEEN 75228160 AND 75235759 OR `expr_annotation_1`.`end` >= 75235759 AND `expr_annotation_1`.`start` <= 75228160)) AND `expression_1`.`value` >= 25.63 AND `expression_1`.`sample_id` IN ('sample_1', 'sample_2') ORDER BY `feature_id` ASC, `sample_id` ASC",
            "SELECT `expression_1`.`feature_id` AS `feature_id`, `expression_1`.`sample_id` AS `sample_id`, `expression_1`.`value` AS `expression`, `expr_annotation_1`.`gene_name` AS `feature_name`, `expr_annotation_1`.`chr` AS `chrom`, `expr_annotation_1`.`start` AS `start` FROM `database_gzky7400vgpyzy621q43gkkf__molecular_expression1_db`.`expression` AS `expression_1` LEFT OUTER JOIN `database_gzky7400vgpyzy621q43gkkf__molecular_expression1_db`.`expr_annotation` AS `expr_annotation_1` ON `expression_1`.`feature_id` = `expr_annotation_1`.`feature_id` WHERE (`expr_annotation_1`.`chr` = '11' AND (`expr_annotation_1`.`end` BETWEEN 8693350 AND 67440200 OR `expr_annotation_1`.`start` BETWEEN 8693350 AND 67440200 OR `expr_annotation_1`.`end` >= 67440200 AND `expr_annotation_1`.`start` <= 8693350) OR `expr_annotation_1`.`chr` = 'X' AND (`expr_annotation_1`.`end` BETWEEN 148500700 AND 148994424 OR `expr_annotation_1`.`start` BETWEEN 148500700 AND 148994424 OR `expr_annotation_1`.`end` >= 148994424 AND `expr_annotation_1`.`start` <= 148500700) OR `expr_annotation_1`.`chr` = '17' AND (`expr_annotation_1`.`end` BETWEEN 75228160 AND 75235759 OR `expr_annotation_1`.`start` BETWEEN 75228160 AND 75235759 OR `expr_annotation_1`.`end` >= 75235759 AND `expr_annotation_1`.`start` <= 75228160)) AND `expression_1`.`sample_id` IN ('sample_1', 'sample_2') AND `expression_1`.`value` >= 25.63 ORDER BY `feature_id` ASC, `sample_id` ASC",
        ]
        response = self.run_dx_extract_assay_expression_cmd(
            self.expression_dataset,
            EXPRESSION_CLI_JSON_FILTERS["positive_test"]["location_expression_sample"],
            "chrom,start,feature_name",
            True,
            "-",
        )
        self.assertIn(response.strip(), expected_sql_query)

    def test_dx_extract_cmd_location_expression_sample_data(self):
        response = self.run_dx_extract_assay_expression_cmd(
            self.expression_dataset,
            EXPRESSION_CLI_JSON_FILTERS["positive_test"]["location_expression_sample"],
            "chrom,start,feature_name",
            False,
            "-",
        )
        response = response.splitlines()
        response_list = [s.split(",") for s in response]
        column_names = response_list[0]
        response_df = pd.DataFrame(response_list[1:], columns=column_names)

        expected_present_row = response_df.loc[
            (response_df["feature_id"] == "ENST00000683201")
            & (response_df["expression"] == "27")
            & (response_df["start"] == "57805541")
            & (response_df["chrom"] == "11")
            & (response_df["feature_name"] == "CTNND1")
            & (response_df["sample_id"] == "sample_2")
        ]

        expected_X_chrom_response = response_df.loc[
            (response_df["chrom"] == "X") & (response_df["sample_id"] == "sample_2")
        ]

        self.assertEqual(len(response_df), 9398)
        self.assertEqual(
            set(column_names),
            set(
                [
                    "feature_id",
                    "sample_id",
                    "expression",
                    "feature_name",
                    "chrom",
                    "start",
                ]
            ),
        )
        self.assertEqual(set(response_df.sample_id), set(["sample_1", "sample_2"]))
        self.assertEqual(len(set(response_df.feature_id.unique())), 5929)
        self.assertEqual(len(expected_present_row), 1)
        self.assertEqual(len(expected_X_chrom_response), 5)

    def test_dx_extract_cmd_with_additional_fields(self):
        expected_sql_query_elements = [
            "SELECT `expression_1`.`feature_id` AS `feature_id`, "
            "`expression_1`.`sample_id` AS `sample_id`,",
            "`expression_1`.`value` AS `expression`,",
            "`expr_annotation_1`.`gene_name` AS `feature_name`,",
            "`expr_annotation_1`.`chr` AS `chrom`,",
            "`expr_annotation_1`.`start` AS `start`,",
            "`expr_annotation_1`.`end` AS `end`,",
            "`expr_annotation_1`.`strand` AS `strand`",
            "FROM `database_gzky7400vgpyzy621q43gkkf__molecular_expression1_db`.`expression` AS `expression_1`",
            "LEFT OUTER JOIN `database_gzky7400vgpyzy621q43gkkf__molecular_expression1_db`.`expr_annotation` AS `expr_annotation_1` ON `expression_1`.`feature_id` = `expr_annotation_1`.`feature_id` "
            "WHERE",
            "`expr_annotation_1`.`feature_id` IN ('ENST00000327669', 'ENST00000456328')",
            "AND",
            "`expression_1`.`sample_id` IN ('sample_1')",
            "ORDER BY `feature_id` ASC, `sample_id` ASC",
        ]
        response = self.run_dx_extract_assay_expression_cmd(
            self.expression_dataset,
            EXPRESSION_CLI_JSON_FILTERS["positive_test"]["additional_fields"],
            "chrom,start,end,strand,feature_name",
            True,
            "-",
        )
        for e in expected_sql_query_elements:
            self.assertIn(e, response)

    def test_negative_dx_extract_cmd_empty_json(self):
        expected_error = "No filter JSON is passed with --retrieve-expression or input JSON for --retrieve-expression does not contain valid filter information."
        response = self.run_dx_extract_assay_expression_cmd(
            self.expression_dataset,
            EXPRESSION_CLI_JSON_FILTERS["negative_test"]["empty_json"],
            None,
            True,
            "-",
            subprocess_run=True,
        )
        self.assertIn(expected_error, response.stderr)

    def test_negative_dx_extract_cmd_invalid_location_range(self):
        expected_error = "Range cannot be greater than 250000000 for location"
        response = self.run_dx_extract_assay_expression_cmd(
            self.expression_dataset,
            EXPRESSION_CLI_JSON_FILTERS["negative_test"]["large_location_range"],
            None,
            False,
            subprocess_run=True,
        )
        self.assertIn(expected_error, response.stderr)

    def test_negative_dx_extract_cmd_too_many_sample_ids(self):
        expected_error = "Too many items given in field sample_id, maximum is 100"
        response = self.run_dx_extract_assay_expression_cmd(
            self.expression_dataset,
            EXPRESSION_CLI_JSON_FILTERS["negative_test"]["sample_id_maxitem_limit"],
            None,
            False,
            subprocess_run=True,
        )
        self.assertIn(expected_error, response.stderr)

    def test_merge_duplicate_filters_schema_1_0(self):
        json_input = VIZPAYLOADERBUILDER_TEST_INPUT[
            "test_vizpayloadbuilder_expression_max"
        ]
        schema = EXTRACT_ASSAY_EXPRESSION_FILTERING_CONDITIONS_1_0
        input_json_parser = JSONFiltersValidator(json_input, schema)

        raw_compound_filter = input_json_parser.parse_v1()

        # For context, in this case:
        # raw_compound_filter = {
        #     "logic": "and",
        #     "compound": [
        #         {
        #             "filters": {
        #                 "expression$value": [{"condition": "less-than-eq", "values": 10}]
        #             }
        #         },
        #         {
        #             "filters": {
        #                 "expr_annotation$gene_name": [{"condition": "in", "values": ["BRCA2"]}]
        #             }
        #         },
        #     ],
        # }

        expected_merged_compound_block = {
            "logic": "and",
            "compound": [
                {
                    "filters": {
                        "expression$value": [
                            {"condition": "less-than-eq", "values": 10}
                        ],
                        "expr_annotation$gene_name": [
                            {"condition": "in", "values": ["BRCA2"]}
                        ],
                    }
                }
            ],
        }

        merged_compound_block = input_json_parser.merge_duplicate_filters(
            raw_compound_filter
        )
        assert merged_compound_block == expected_merged_compound_block

    def test_merge_duplicate_filters_schema_1_1(self):
        json_input = VIZPAYLOADERBUILDER_TEST_INPUT[
            "test_vizpayloadbuilder_expression_max"
        ]
        schema = EXTRACT_ASSAY_EXPRESSION_FILTERING_CONDITIONS_1_1
        input_json_parser = JSONFiltersValidator(json_input, schema)

        raw_compound_filter = input_json_parser.parse_v1()

        # For context, in this case:
        # raw_compound_filter = {
        #     "logic": "and",
        #     "compound": [
        #         {
        #             "filters": {
        #                 "expression_read_optimized$value": [
        #                     {"condition": "less-than-eq", "values": 10}
        #                 ]
        #             }
        #         },
        #         {
        #             "filters": {
        #                 "expression_read_optimized$gene_name": [
        #                     {"condition": "in", "values": ["BRCA2"]}
        #                 ]
        #             }
        #         },
        #     ],
        # }

        expected_merged_compound_block = {
            "logic": "and",
            "compound": [
                {
                    "filters": {
                        "expression_read_optimized$value": [
                            {"condition": "less-than-eq", "values": 10}
                        ],
                        "expression_read_optimized$gene_name": [
                            {"condition": "in", "values": ["BRCA2"]}
                        ],
                    }
                }
            ],
        }

        merged_compound_block = input_json_parser.merge_duplicate_filters(
            raw_compound_filter
        )
        assert merged_compound_block == expected_merged_compound_block

    def test_merge_duplicate_filters_schema_1_1_non_optimized(self):
        json_input = VIZPAYLOADERBUILDER_TEST_INPUT[
            "test_vizpayloadbuilder_expression_max"
        ]
        schema = EXTRACT_ASSAY_EXPRESSION_FILTERING_CONDITIONS_1_1_non_optimized
        input_json_parser = JSONFiltersValidator(json_input, schema)

        raw_compound_filter = input_json_parser.parse_v1()

        # For context, in this case:
        # raw_compound_filter = {
        #     "compound": [
        #         {
        #             "filters": {
        #                 "expression$value": [{"condition": "less-than-eq", "values": 10}]
        #             }
        #         },
        #         {
        #             "filters": {
        #                 "expr_annotation$gene_name": [{"condition": "in", "values": [...]}]
        #             }
        #         },
        #     ],
        #     "logic": "and",
        # }

        expected_merged_compound_block = {
            "logic": "and",
            "compound": [
                {
                    "filters": {
                        "expression$value": [
                            {"condition": "less-than-eq", "values": 10}
                        ],
                        "expr_annotation$gene_name": [
                            {"condition": "in", "values": ["BRCA2"]}
                        ],
                    }
                }
            ],
        }

        merged_compound_block = input_json_parser.merge_duplicate_filters(
            raw_compound_filter
        )
        assert merged_compound_block == expected_merged_compound_block


# Start the test
if __name__ == "__main__":
    unittest.main()



# Copyright (C) 2013-2018 DNAnexus, Inc.
#
# This file is part of dx-toolkit (DNAnexus platform client libraries).
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may not
#   use this file except in compliance with the License. You may obtain a copy
#   of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

"""
Utilities used when running executables in a batch. For example, running an analysis on
a set of data samples.
"""

from __future__ import print_function, unicode_literals, division, absolute_import

from collections import defaultdict
import csv
import dxpy
import json

from ..exceptions import err_exit


# Informational columns in the TSV file, which we want to ignore
# TODO: are these the correct names?
BATCH_ID = "batch ID"

# Figure out the type for each input, by describing the
# executable
def _get_types_for_inputs(executable):
    exec_desc = executable.describe()
    input_spec = []
    if ('inputs' in exec_desc and
        exec_desc['inputs'] is not None):
        # workflow-level inputs were defined for the workflow
        input_spec = exec_desc['inputs']
    elif ('inputSpec' in exec_desc and
          exec_desc['inputSpec'] is not None):
        input_spec = exec_desc['inputSpec']
    else:
        raise Exception("executable {} does not have an input specification".format(executable.get_id()))

    input_key_classes={}
    for arg_desc in input_spec:
        input_key_classes[arg_desc['name']] = arg_desc['class']

    return input_key_classes

# val: a value of type string, that needs to be converted
# col_class: the class of the executable argument
#
# return:
#   - the value in the correct type
#   - any platform files referenced in the type
def _type_convert_primitive(val, klass):
    retval = None
    ref_files = []
    if klass == 'string':
        retval = val
    elif klass == 'int':
        retval = int(val)
    elif klass == "boolean":
        retval = bool(val)
    elif klass == 'float':
        retval = float(val)
    elif klass == 'hash':
        retval = json.loads(val)
    elif klass == 'file':
        if val.startswith("project-"):
            val = val.split(":")
            retval = dxpy.dxlink(object_id=val[1], project_id=val[0])
            ref_files.append(retval)
        elif val.startswith("file-"):
            retval = dxpy.dxlink(val)
            ref_files.append(retval)
        else:
            raise Exception("Malformed file {}, must start with 'file-' or 'project-'".format(val))
    else:
        raise Exception("class {} not currently supported".format(klass))
    return retval, ref_files

# An array type, such as array:file
def _type_convert_array(val, klass):
    inner_type = klass.split(":")[1]
    len_val = len(val)
    if (len_val < 2 or
        val[0] != '[' or
        val[len_val-1] != ']'):
        raise Exception("Malformed array {}".format(val))
    val_strip_brackets = val[1:(len_val-1)]
    retval = []
    ref_files = []
    elements = [e.strip() for e in val_strip_brackets.split(',')]
    for e in elements:
        e_with_type,files = _type_convert_primitive(e, inner_type)
        retval.append(e_with_type)
        ref_files += files
    return retval, ref_files

def _type_convert(val, klass):
    try:
        if klass.startswith("array:"):
            return _type_convert_array(val, klass)
        else:
            # A primitive type, like {int, float, file, hash}
            return _type_convert_primitive(val, klass)
    except Exception:
        raise Exception("value={} cannot be converted into class {}".format(val, klass))

# For a column that represents files, assume it is named "pair", look for
# the index of column "pair ID".
def _search_column_id(col_name, header_line):
    for i, col_name2 in enumerate(header_line):
        if col_name2 == (col_name + " ID"):
            return i
    raise Exception("Could not find a column with file IDs for {}".format(col_name))

# Parse the TSV file. Create a dictionary with the input arguments for
# each invocation. Return an array of dictionaries.
#
# For each line, except the header:
#   remove informational columns
#   create a dictionary of inputs we can pass to the executable
#
# Example TSV input:
# batch ID, pair1,       pair1 ID, pair2,       pair2 ID
# 23,       SRR123_1.gz, file-XXX, SRR223_2.gz, file-YYY
#
def batch_launch_args(executable, input_json, batch_tsv_file):
    header_line = []
    lines = []
    read_mode = "r"
    with open(batch_tsv_file, read_mode) as f:
        reader = csv.reader(f, delimiter=str(u'\t'))
        header_line = next(reader)
        lines = list(reader)

    # Get the classes for the executable inputs
    input_classes = _get_types_for_inputs(executable)

    # which columns to use. Note that a file input uses two columns.
    # For example, argument 'pair1' exposes two columns {'pair1, 'pair1 ID'}.
    batch_index = None
    index_2_column = {}
    for i, col_name in enumerate(header_line):
        if col_name == BATCH_ID:
            batch_index = i
        if col_name not in input_classes:
            continue
        if input_classes[col_name] in ['file', 'array:file']:
            idx = _search_column_id(col_name, header_line)
            index_2_column[idx] = col_name
        else:
            index_2_column[i] = col_name
    if batch_index is None:
        raise Exception("Could not find column {}".format(BATCH_ID))

    # A dictionary of inputs. Each column in the TSV file is mapped to a row.
    # {
    #    "a": [{dnanexus_link: "file-xxxx"}, {dnanexus_link: "file-yyyy"}, ....],
    #    "b": [1,null, ...]
    # }
    columns=defaultdict(list)
    all_files=[]
    for line in lines:
        for i, val in enumerate(line):
            if i not in index_2_column:
                continue
            col_name = index_2_column[i]
            klass = input_classes[col_name]
            val_w_correct_type, ref_files = _type_convert(val.strip(), klass)
            all_files += ref_files
            columns[col_name].append(val_w_correct_type)

    # Create an array of batch_ids
    batch_ids = [line[batch_index].strip() for line in lines]

    # call validate
    #
    # Output: list of dictionaries, each dictionary corresponds to expanded batch call
    expanded_args = dxpy.api.applet_validate_batch(executable.get_id(),
                                                   { "batchInput": columns,
                                                     "commonInput": input_json,
                                                     "files": all_files,
                                                     "instanceTypes": [] })

    ## future proofing
    if isinstance(expanded_args, dict):
        assert('expandedBatch' in expanded_args)
        launch_args = expanded_args['expandedBatch']
    else:
        launch_args = expanded_args
    if len(launch_args) != len(batch_ids):
        raise Exception("Mismatch in number of launch_args vs. batch_ids ({} != {})"
                        .format(len(launch_args), len(batch_ids)))

    return { "launch_args": launch_args,
             "batch_ids": batch_ids }

#
# executable: applet, app, or workflow
# launch_args: array of dictionaries, each of which contains all arguments needed to
#     invoke the executable.
# set_batch_folders: boolean, if True, the results from each batch
#     run will be placed in a separate output folder named after batch ID.
#     The folders will be created/used relative to the output folder set with
#     the --destination arg
def batch_run(executable, b_args, run_kwargs, set_batch_folders=False):
    run_args = run_kwargs.copy()
    exec_name = executable.describe()["name"]
    launch_args = b_args["launch_args"]
    batch_ids = b_args["batch_ids"]
    executions = []
    for idx, input_json in enumerate(launch_args):
        batch_id = batch_ids[idx]
        name = "{}-{}".format(exec_name, batch_id)
        properties = {
            'batch-id': batch_id,
            'batch-name': name
        }
        run_args['name'] = name
        if run_args.get('properties') is not None:
            run_args['properties'].update(properties)
        else:
            run_args['properties'] = properties

        if set_batch_folders:
            run_args['folder'] = run_kwargs['folder'] + "/" + batch_id

        try:
            dxexecution = executable.run(input_json, **run_args)
            executions.append(dxexecution)
        except Exception:
            err_exit()
    return executions



# Copyright (C) 2013-2016 DNAnexus, Inc.
#
# This file is part of dx-toolkit (DNAnexus platform client libraries).
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may not
#   use this file except in compliance with the License. You may obtain a copy
#   of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import re

_cache = {}
_MAXCACHE = 100

def filter(names, pat):
    """
    A copy of fnmatch.filter that replaces with [^\/] instead of . to match path globs.
    """
    import os,posixpath
    result=[]
    pat=os.path.normcase(pat)
    if not pat in _cache:
        res = translate(pat)
        if len(_cache) >= _MAXCACHE:
            _cache.clear()
        _cache[pat] = re.compile(res)
    match=_cache[pat].match
    if os.path is posixpath:
        # normcase on posix is NOP. Optimize it away from the loop.
        for name in names:
            if match(name):
                result.append(name)
    else:
        for name in names:
            if match(os.path.normcase(name)):
                result.append(name)
    return result

def translate(pat):
    """
    A copy of fnmatch.translate that replaces with [^\/] instead of . to match path globs.
    """

    i, n = 0, len(pat)
    res = ''
    while i < n:
        c = pat[i]
        i = i+1
        if c == '*':
            res = res + '[^\/]*'
        elif c == '?':
            res = res + '[^\/].'
        elif c == '[':
            j = i
            if j < n and pat[j] == '!':
                j = j+1
            if j < n and pat[j] == ']':
                j = j+1
            while j < n and pat[j] != ']':
                j = j+1
            if j >= n:
                res = res + '\\['
            else:
                stuff = pat[i:j].replace('\\','\\\\')
                i = j+1
                if stuff[0] == '!':
                    stuff = '^' + stuff[1:]
                elif stuff[0] == '^':
                    stuff = '\\' + stuff
                res = '%s[%s]' % (res, stuff)
        else:
            res = res + re.escape(c)
    return '(?ms)' + res + '\Z'



#!/usr/bin/env python3
#
# Copyright (C) 2013-2016 DNAnexus, Inc.
#
# This file is part of dx-toolkit (DNAnexus platform client libraries).
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may not
#   use this file except in compliance with the License. You may obtain a copy
#   of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.


from __future__ import print_function, unicode_literals, division, absolute_import

import logging
logging.basicConfig(level=logging.WARNING)
logging.getLogger('urllib3.connectionpool').setLevel(logging.ERROR)

import os, sys, json, subprocess, argparse
import platform
import py_compile
import re
import shutil
import tempfile
import time
from datetime import datetime
import dxpy
import dxpy.app_builder
import dxpy.workflow_builder
import dxpy.executable_builder
from .. import logger

from dxpy.nextflow.nextflow_builder import build_pipeline_with_npi, prepare_nextflow
from dxpy.nextflow.nextflow_utils import get_resources_subpath, is_importer_job

from ..utils import json_load_raise_on_duplicates
from ..utils.resolver import resolve_path, check_folder_exists, ResolutionError, is_container_id
from ..utils.completer import LocalCompleter
from ..app_categories import APP_CATEGORIES
from ..exceptions import err_exit
from ..utils.printing import BOLD
from ..compat import USING_PYTHON2, basestring
from ..cli.parsers import process_extra_args


parser = argparse.ArgumentParser(description="Uploads a DNAnexus App.")

class DXSyntaxError(Exception):
    def __init__(self, message):
        self.message = message
    def __str__(self):
        return self.message

def _get_timestamp_version_suffix(version):
    if "+" in version:
        return ".build." + datetime.today().strftime('%Y%m%d.%H%M')
    else:
        return "+build." + datetime.today().strftime('%Y%m%d.%H%M')

def _get_version_suffix(src_dir, version):
    # If anything goes wrong, fall back to the date-based suffix.
    try:
        if os.path.exists(os.path.join(src_dir, ".git")):
            abbrev_sha1 = subprocess.check_output(['git', 'rev-parse', 'HEAD'], cwd=src_dir).strip()[:7]
            # We ensure that if VERSION is semver-compliant, then
            # VERSION + SUFFIX will be too. In particular that means
            # (here and in _get_timestamp_version_suffix above) we add
            # what semver refers to as a "build metadata" section
            # (delimited by "+"), unless one already exists, in which
            # case we append to the existing one.
            if "+" in version:
                return ".git." + abbrev_sha1
            else:
                return "+git." + abbrev_sha1
    except:
        pass
    return _get_timestamp_version_suffix(version)

def parse_destination(dest_str):
    return dxpy.executable_builder.get_parsed_destination(dest_str)

def _check_suggestions(app_json, publish=False):
    """
    Examines the specified dxapp.json file and warns about any
    violations of suggestions guidelines.

    :raises: AppBuilderException for data objects that could not be found
    """
    for input_field in app_json.get('inputSpec', []):
        for suggestion in input_field.get('suggestions', []):
            if 'project' in suggestion:
                try:
                    project = dxpy.api.project_describe(suggestion['project'], {"permissions": True})
                    if 'PUBLIC' not in project['permissions'] and publish:
                        logger.warn('Project {name} NOT PUBLIC!'.format(name=project['name']))
                except dxpy.exceptions.DXAPIError as e:
                    if e.code == 404:
                        logger.warn('Suggested project {name} does not exist, or not accessible by user'.format(
                                     name=suggestion['project']))
                if 'path' in suggestion:
                    try:
                        check_folder_exists(suggestion['project'], suggestion['path'], '')
                    except ResolutionError as e:
                        logger.warn('Folder {path} could not be found in project {project}'.format(
                                     path=suggestion['path'], project=suggestion['project']))
            if '$dnanexus_link' in suggestion:
                if suggestion['$dnanexus_link'].startswith(('file-', 'record-')):
                    try:
                        dnanexus_link = dxpy.describe(suggestion['$dnanexus_link'])
                    except dxpy.exceptions.DXAPIError as e:
                        if e.code == 404:
                            raise dxpy.app_builder.AppBuilderException(
                                'Suggested object {name} could not be found'.format(
                                    name=suggestion['$dnanexus_link']))
                    except Exception as e:
                        raise dxpy.app_builder.AppBuilderException(str(e))
            if 'value' in suggestion and isinstance(suggestion["value"], dict):
                if '$dnanexus_link' in suggestion['value']:
                    # Check if we have JSON or string
                    if isinstance(suggestion['value']['$dnanexus_link'], dict):
                        if 'project' in suggestion['value']['$dnanexus_link']:
                            try:
                                dxpy.api.project_describe(suggestion['value']['$dnanexus_link']['project'])
                            except dxpy.exceptions.DXAPIError as e:
                                if e.code == 404:
                                    logger.warn('Suggested project {name} does not exist, or not accessible by user'.format(
                                                 name=suggestion['value']['$dnanexus_link']['project']))
                    elif isinstance(suggestion['value']['$dnanexus_link'], basestring):
                        if suggestion['value']['$dnanexus_link'].startswith(('file-', 'record-')):
                            try:
                                dnanexus_link = dxpy.describe(suggestion['value']['$dnanexus_link'])
                            except dxpy.exceptions.DXAPIError as e:
                                if e.code == 404:
                                    raise dxpy.app_builder.AppBuilderException(
                                        'Suggested object {name} could not be found'.format(
                                            name=suggestion['value']['$dnanexus_link']))
                            except Exception as e:
                                raise dxpy.app_builder.AppBuilderException(str(e))

def _lint(dxapp_json_filename, mode):
    """
    Examines the specified dxapp.json file and warns about any
    violations of app guidelines.

    Precondition: the dxapp.json file exists and can be parsed.
    """

    def _find_readme(dirname):
        for basename in ['README.md', 'Readme.md', 'readme.md']:
            if os.path.exists(os.path.join(dirname, basename)):
                return os.path.join(dirname, basename)
        return None

    app_spec = json.load(open(dxapp_json_filename))

    dirname = os.path.basename(os.path.dirname(os.path.abspath(dxapp_json_filename)))

    if mode == "app":
        if 'title' not in app_spec:
            logger.warn('app is missing a title, please add one in the "title" field of dxapp.json')

        if 'summary' in app_spec:
            if app_spec['summary'].endswith('.'):
                logger.warn('summary "%s" should be a short phrase not ending in a period' % (app_spec['summary'],))
        else:
            logger.warn('app is missing a summary, please add one in the "summary" field of dxapp.json')

        readme_filename = _find_readme(os.path.dirname(dxapp_json_filename))
        if 'description' in app_spec:
            if readme_filename:
                raise dxpy.app_builder.AppBuilderException('Description was provided both in Readme.md '
                          'and in the "description" field of {file}. Please consolidate content in Readme.md '
                          'and remove the "description" field.'.format(file=dxapp_json_filename))
            if not app_spec['description'].strip().endswith('.'):
                logger.warn('"description" field should be written in complete sentences and end with a period')
        else:
            if readme_filename is None:
                logger.warn("app is missing a description, please supply one in README.md")
        if 'categories' in app_spec:
            for category in app_spec['categories']:
                if category not in APP_CATEGORIES:
                    logger.warn('app has unrecognized category "%s"' % (category,))
                if category == 'Import':
                    if 'title' in app_spec and not app_spec['title'].endswith('Importer'):
                        logger.warn('title "%s" should end in "Importer"' % (app_spec['title'],))
                if category == 'Export':
                    if 'title' in app_spec and not app_spec['title'].endswith('Exporter'):
                        logger.warn('title "%s" should end in "Exporter"' % (app_spec['title'],))

    if 'name' in app_spec:
        if app_spec['name'] != app_spec['name'].lower():
            logger.warn('name "%s" should be all lowercase' % (app_spec['name'],))
        if dirname != app_spec['name'] and not os.path.abspath(dxapp_json_filename).startswith("/tmp") and not dirname.startswith("."):
            logger.warn('app name "%s" does not match containing directory "%s"' % (app_spec['name'], dirname))
    else:
        logger.warn('app is missing a name, please add one in the "name" field of dxapp.json')

    if 'version' in app_spec:
        if not dxpy.executable_builder.GLOBAL_EXEC_VERSION_RE.match(app_spec['version']):
            logger.warn('"version" %s should be semver compliant (e.g. of the form X.Y.Z)' % (app_spec['version'],))

    # Note that identical checks are performed on the server side (and
    # will cause the app build to fail), but the checks here are printed
    # sooner and multiple naming problems can be detected in a single
    # pass.
    if 'inputSpec' in app_spec:
        for i, input_field in enumerate(app_spec['inputSpec']):
            if not re.match("^[a-zA-Z_][0-9a-zA-Z_]*$", input_field['name']):
                logger.error('input %d has illegal name "%s" (must match ^[a-zA-Z_][0-9a-zA-Z_]*$)' % (i, input_field['name']))
    else:
        logger.warn("dxapp.json contains no input specification (inputSpec). Your applet will not be usable as an " +
                    "app, runnable from the GUI, or composable using workflows.")
    if 'outputSpec' in app_spec:
        for i, output_field in enumerate(app_spec['outputSpec']):
            if not re.match("^[a-zA-Z_][0-9a-zA-Z_]*$", output_field['name']):
                logger.error('output %d has illegal name "%s" (must match ^[a-zA-Z_][0-9a-zA-Z_]*$)' % (i, output_field['name']))
    else:
        logger.warn("dxapp.json contains no output specification (outputSpec). Your applet will not be usable as an " +
                    "app, runnable from the GUI, or composable using workflows.")

def _check_syntax(code, lang, temp_dir, enforce=True):
    """
    Checks that the code whose text is in CODE parses as LANG.

    Raises DXSyntaxError if there is a problem and "enforce" is True.
    """
    # This function needs the language to be explicitly set, so we can
    # generate an appropriate temp filename.
    if lang == 'python2.7' or 'python3' in lang:
        temp_basename = 'inlined_code_from_dxapp_json.py'
    elif lang == 'bash':
        temp_basename = 'inlined_code_from_dxapp_json.sh'
    else:
        raise ValueError('lang must be one of "python2.7", "python3", or "bash"')
    # Dump the contents out to a temporary file, then call _check_file_syntax.
    with open(os.path.join(temp_dir, temp_basename), 'w') as ofile:
        ofile.write(code)
    _check_file_syntax(os.path.join(temp_dir, temp_basename), temp_dir, override_lang=lang, enforce=enforce)


# convert the error message embedded in the exception to a proper string
def _error_message_to_string(e, message):
    if isinstance(message, str):
        return message
    elif isinstance(message, bytes):
        return message.decode("utf-8")
    else:
        # What kind of object is this?
        print("The error message is neither string nor bytes, it is {}".format(type(message)))
        raise e

def _check_file_syntax(filename, temp_dir, override_lang=None, enforce=True):
    """
    Checks that the code in FILENAME parses, attempting to autodetect
    the language if necessary.

    Raises IOError if the file cannot be read.

    Raises DXSyntaxError if there is a problem and "enforce" is True.
    """
    def check_python(filename):
        # Generate a semi-recognizable name to write the pyc to. Of
        # course it's possible that different files being scanned could
        # have the same basename, so this path won't be unique, but the
        # checks don't run concurrently so this shouldn't cause any
        # problems.
        pyc_path = os.path.join(temp_dir, os.path.basename(filename) + ".pyc")
        try:
            py_compile.compile(filename, cfile=pyc_path, doraise=True)
        finally:
            try:
                os.unlink(pyc_path)
            except OSError:
                pass
    def check_bash(filename):
        if platform.system() == 'Windows':
            logging.warn(
                    'Skipping bash syntax check due to unavailability of bash on Windows.')
        else:
            subprocess.check_output(["/bin/bash", "-n", filename], stderr=subprocess.STDOUT)
    python_unsure = False
    if override_lang == 'python2.7' and USING_PYTHON2:
        checker_fn = check_python
    elif override_lang == 'python3' and not USING_PYTHON2:
        checker_fn = check_python
    elif override_lang == 'bash':
        checker_fn = check_bash
    elif filename.endswith('.py'):
        checker_fn = check_python
        # don't enforce and ignore if the shebang is ambiguous and we're not sure
        # that the file version is the same as the one we're running
        read_mode = "r"
        with open(filename, read_mode) as f:
            first_line = f.readline()
            if not (('python3' in first_line and not USING_PYTHON2) or
                    ('python2' in first_line and USING_PYTHON2)):
                enforce = False
                python_unsure = True
    elif filename.endswith('.sh'):
        checker_fn = check_bash
    else:
        # Ignore other kinds of files.
        return

    # Do a test read of the file to catch errors like the file not
    # existing or not being readable.
    open(filename)

    try:
        checker_fn(filename)
    except subprocess.CalledProcessError as e:
        print(filename + " has a syntax error! Interpreter output:", file=sys.stderr)
        errmsg = _error_message_to_string(e, e.output)
        for line in errmsg.strip("\n").split("\n"):
            print("  " + line.rstrip("\n"), file=sys.stderr)
        if enforce:
            raise DXSyntaxError(filename + " has a syntax error")
    except py_compile.PyCompileError as e:
        if python_unsure:
            print("Unsure if " + filename + " is using Python 2 or Python 3, the following error might not be relevant", file=sys.stderr)
        print(filename + " has a syntax error! Interpreter output:", file=sys.stderr)
        errmsg = _error_message_to_string(e, e.msg)
        print("  " + errmsg.strip(), file=sys.stderr)
        if enforce:
            raise DXSyntaxError(e.msg.strip())


def _verify_app_source_dir_impl(src_dir, temp_dir, mode, enforce=True):
    """Performs syntax and lint checks on the app source.

    Precondition: the dxapp.json file exists and can be parsed.
    """
    _lint(os.path.join(src_dir, "dxapp.json"), mode)

    # Check that the entry point file parses as the type it is going to
    # be interpreted as. The extension is irrelevant.
    manifest = json.load(open(os.path.join(src_dir, "dxapp.json")))
    if "runSpec" in manifest:
        if "interpreter" not in manifest['runSpec']:
            raise dxpy.app_builder.AppBuilderException('runSpec.interpreter field was not present')

        if "distribution" not in manifest['runSpec']:
            raise dxpy.app_builder.AppBuilderException('Required field runSpec.distribution is not present')

        if "release" not in manifest['runSpec']:
            raise dxpy.app_builder.AppBuilderException('Required field runSpec.release is not present')

        if manifest['runSpec']['interpreter'] in ["python2.7", "bash", "python3"]:
            if "file" in manifest['runSpec']:
                entry_point_file = os.path.abspath(os.path.join(src_dir, manifest['runSpec']['file']))
                try:
                    _check_file_syntax(entry_point_file, temp_dir, override_lang=manifest['runSpec']['interpreter'], enforce=enforce)
                except IOError as e:
                    raise dxpy.app_builder.AppBuilderException(
                        'Could not open runSpec.file=%r. The problem was: %s' % (entry_point_file, e))
                except DXSyntaxError:
                    raise dxpy.app_builder.AppBuilderException('Entry point file %s has syntax errors, see above for details. Rerun with --no-check-syntax to proceed anyway.' % (entry_point_file,))
            elif "code" in manifest['runSpec']:
                try:
                    _check_syntax(manifest['runSpec']['code'], manifest['runSpec']['interpreter'], temp_dir, enforce=enforce)
                except DXSyntaxError:
                    raise dxpy.app_builder.AppBuilderException('Code in runSpec.code has syntax errors, see above for details. Rerun with --no-check-syntax to proceed anyway.')

            cluster_bootstrap_scripts = _get_all_cluster_bootstrap_script_names(manifest)
            for filename in cluster_bootstrap_scripts:
                abs_filename = os.path.abspath(os.path.join(src_dir, filename))
                try:
                    _check_file_syntax(abs_filename,
                                       temp_dir,
                                       override_lang=manifest['runSpec']['interpreter'],
                                       enforce=enforce)
                except IOError as e:
                    raise dxpy.app_builder.AppBuilderException(
                        'Could not open cluster bootstrap script %r. The problem was: %s' % (abs_filename, e))
                except DXSyntaxError:
                    raise dxpy.app_builder.AppBuilderException('Code in cluster bootstrapScript %r has syntax errors, see above for details. Rerun with --no-check-syntax to proceed anyway.' % filename)

        if 'execDepends' in manifest['runSpec']:
            if not isinstance(manifest['runSpec']['execDepends'], list):
                raise dxpy.app_builder.AppBuilderException('Expected runSpec.execDepends to be an array. Rerun with --no-check-syntax to proceed anyway.')
            if not all(isinstance(dep, dict) for dep in manifest['runSpec']['execDepends']):
                raise dxpy.app_builder.AppBuilderException('Expected runSpec.execDepends to be an array of hashes. Rerun with --no-check-syntax to proceed anyway.')
            if any(dep.get('package_manager', 'apt') != 'apt' for dep in manifest['runSpec']['execDepends']):
                if not isinstance(manifest.get('access'), dict) or 'network' not in manifest['access']:
                    msg = '\n'.join(['runSpec.execDepends specifies non-APT dependencies, but no network access spec is given.',
                    'Add {"access": {"network": ["*"]}} to allow dependencies to install.',
                    'See https://documentation.dnanexus.com/developer/apps/execution-environment#network-access.',
                    'Rerun with --no-check-syntax to proceed anyway.'])
                    raise dxpy.app_builder.AppBuilderException(msg)

    if 'authorizedUsers' in manifest:
        if not isinstance(manifest['authorizedUsers'], list) or isinstance(manifest['authorizedUsers'], basestring):
            raise dxpy.app_builder.AppBuilderException('Expected authorizedUsers to be a list of strings')
        for thing in manifest['authorizedUsers']:
            if thing != 'PUBLIC' and (not isinstance(thing, basestring) or not re.match("^(org-|user-)", thing)):
                raise dxpy.app_builder.AppBuilderException('authorizedUsers field contains an entry which is not either the string "PUBLIC" or a user or org ID')

    if "pricingPolicy" in manifest:
        error_message = "\"pricingPolicy\" at the top level is not accepted. It must be specified "
        error_message += "under the \"regionalOptions\" field in all enabled regions of the app"
        raise dxpy.app_builder.AppBuilderException(error_message)

    # Check all other files that are going to be in the resources tree.
    # For these we detect the language based on the filename extension.
    # Obviously this check can have false positives, since the app can
    # execute (or not execute!) all these files in whatever way it
    # wishes, e.g. it could use Python != 2.7 or some non-bash shell.
    # Consequently errors here are non-fatal.
    files_with_problems = []
    for dirpath, dirnames, filenames in os.walk(os.path.abspath(os.path.join(src_dir, "resources"))):
        for filename in filenames:
            # On Mac OS, the resource fork for "FILE.EXT" gets tarred up
            # as a file named "._FILE.EXT". To a naive check this
            # appears to be a file of the same extension. Therefore, we
            # exclude these from syntax checking since they are likely
            # to not parse as whatever language they appear to be.
            if not filename.startswith("._"):
                try:
                    _check_file_syntax(os.path.join(dirpath, filename), temp_dir, enforce=True)
                except IOError as e:
                    raise dxpy.app_builder.AppBuilderException(
                        'Could not open file in resources directory %r. The problem was: %s' %
                        (os.path.join(dirpath, filename), e)
                    )
                except DXSyntaxError:
                    # Suppresses errors from _check_file_syntax so we
                    # only print a nice error message
                    files_with_problems.append(os.path.join(dirpath, filename))

    if files_with_problems:
        # Make a message of the form:
        #    "/path/to/my/app.py"
        # OR "/path/to/my/app.py and 3 other files"
        files_str = files_with_problems[0] if len(files_with_problems) == 1 else (files_with_problems[0] + " and " + str(len(files_with_problems) - 1) + " other file" + ("s" if len(files_with_problems) > 2 else ""))
        logging.warn('%s contained syntax errors, see above for details' % (files_str,))


def _get_all_cluster_bootstrap_script_names(manifest):
    # runSpec.systemRequirements is deprecated. Enforce use of regionalOptions.
    if 'systemRequirements' in manifest['runSpec']:
        sys_reqs = manifest['runSpec']['systemRequirements']
        for entry_point in sys_reqs:
            if 'clusterSpec' in sys_reqs[entry_point]:
                err_msg = "\"clusterSpec\" in \"runSpec.systemRequirements\" is not accepted."
                err_msg += " It must be specified in \"systemRequirements\" under the \"regionalOptions\" field in all enabled regions of the app."
                raise dxpy.app_builder.AppBuilderException(err_msg)

    script_names = []
    if 'regionalOptions' in manifest:
        for region in manifest['regionalOptions']:
            if 'systemRequirements' in manifest['regionalOptions'][region]:
                sys_reqs = manifest['regionalOptions'][region]['systemRequirements']
                for entry_point in sys_reqs:
                    try:
                        filename = sys_reqs[entry_point]['clusterSpec']['bootstrapScript']
                        script_names.append(filename)
                    except KeyError:
                        # either no "clusterSpec" or no "bootstrapScript" within "clusterSpec"
                        continue

    return script_names

def _verify_app_source_dir(src_dir, mode, enforce=True):
    """Performs syntax and lint checks on the app source.

    Precondition: the dxapp.json file exists and can be parsed.
    """
    temp_dir = tempfile.mkdtemp(prefix='dx-build_tmp')
    try:
        _verify_app_source_dir_impl(src_dir, temp_dir, mode, enforce=enforce)
    finally:
        shutil.rmtree(temp_dir)

def _parse_app_spec(src_dir):
    """Returns the parsed contents of dxapp.json.

    Raises either AppBuilderException or a parser error (exit codes 3 or
    2 respectively) if this cannot be done.
    """
    if not os.path.isdir(src_dir):
        parser.error("%s is not a directory" % src_dir)
    if not os.path.exists(os.path.join(src_dir, "dxapp.json")):
        raise dxpy.app_builder.AppBuilderException("Directory %s does not contain dxapp.json: not a valid DNAnexus app source directory" % src_dir)
    with open(os.path.join(src_dir, "dxapp.json")) as app_desc:
        try:
            return json_load_raise_on_duplicates(app_desc)
        except Exception as e:
            raise dxpy.app_builder.AppBuilderException("Could not parse dxapp.json file as JSON: " + str(e.args))

def _build_app_remote(mode, src_dir, publish=False, destination_override=None,
                      version_override=None, bill_to_override=None,
                      do_version_autonumbering=True, do_try_update=True, do_parallel_build=True,
                      do_check_syntax=True, region=None, watch=True):
    if mode == 'app':
        builder_app = 'app-tarball_app_builder'
    else:
        builder_app = 'app-tarball_applet_builder'

    app_spec = _parse_app_spec(src_dir)
    builder_versions = {"12.04": "", "14.04": "_trusty", "16.04": "_xenial",
                        "20.04": "_focal", "24.04": "_noble"}
    release = app_spec['runSpec'].get('release')
    # Remote builder app/applet for 16.04 version 1
    if release == "16.04" and app_spec['runSpec'].get('version', '0') == '1':
        builder_app += "_xenial_v1"
    else:
        builder_app += builder_versions.get(release, "")


    temp_dir = tempfile.mkdtemp()

    build_options = {}

    if version_override:
        build_options['version_override'] = version_override
    elif do_version_autonumbering:
        # If autonumbering is DISABLED, the interior run of dx-build-app
        # will detect the correct version to use without our help. If it
        # is ENABLED, the version suffix might depend on the state of
        # the git repository. Since we'll remove the .git directory
        # before uploading, we need to determine the correct version to
        # use here and pass it in to the interior run of dx-build-app.
        if do_version_autonumbering:
            original_version = app_spec['version']
            app_describe = None
            try:
                app_describe = dxpy.api.app_describe("app-" + app_spec["name"], alias=original_version, always_retry=False)
            except dxpy.exceptions.DXAPIError as e:
                if e.name == 'ResourceNotFound' or (mode == 'applet' and e.name == 'PermissionDenied'):
                    pass
                else:
                    raise e
            if app_describe is not None:
                if 'published' in app_describe or not do_try_update:
                    # The version we wanted was taken; fall back to the
                    # autogenerated version number.
                    build_options['version_override'] = original_version + _get_version_suffix(src_dir, original_version)

    # The following flags are basically passed through verbatim.
    if bill_to_override:
        build_options['bill_to_override'] = bill_to_override
    if not do_version_autonumbering:
        build_options['do_version_autonumbering'] = False
    if not do_try_update:
        build_options['do_try_update'] = False
    if not do_parallel_build:
        build_options['do_parallel_build'] = False
    if not do_check_syntax:
        build_options['do_check_syntax'] = False

    using_temp_project_for_remote_build = False

    # If building an applet, run the builder app in the destination
    # project. If building an app, run the builder app in a temporary
    # project.
    dest_folder = None
    dest_applet_name = None
    if mode == "applet":
        # Translate the --destination flag as follows. If --destination
        # is PROJ:FOLDER/NAME,
        #
        # 1. Run the builder app in PROJ
        # 2. Make the output folder FOLDER
        # 3. Supply --destination=NAME to the interior call of dx-build-applet.
        build_project_id = dxpy.WORKSPACE_ID
        if destination_override:
            build_project_id, dest_folder, dest_applet_name = parse_destination(destination_override)
        if build_project_id is None:
            parser.error("Can't create an applet without specifying a destination project; please use the -d/--destination flag to explicitly specify a project")
        if dest_applet_name:
            build_options['destination_override'] = '/' + dest_applet_name

    elif mode == "app":
        using_temp_project_for_remote_build = True
        try:
            project_input = {}
            project_input["name"] = "dx-build-app --remote temporary project"
            if bill_to_override:
                project_input["billTo"] = bill_to_override
            if region:
                project_input["region"] = region
            build_project_id = dxpy.api.project_new(project_input)["id"]
        except:
            err_exit()

    try:
        # Resolve relative paths and symlinks here so we have something
        # reasonable to write in the job name below.
        src_dir = os.path.realpath(src_dir)

        # Show the user some progress as the tarball is being generated.
        # Hopefully this will help them to understand when their tarball
        # is huge (e.g. the target directory already has a whole bunch
        # of binaries in it) and interrupt before uploading begins.
        app_tarball_file = os.path.join(temp_dir, "app_tarball.tar.gz")
        tar_subprocess = subprocess.Popen(["tar", "-czf", "-", "--exclude", "./.git", "."], cwd=src_dir, stdout=subprocess.PIPE)
        with open(app_tarball_file, 'wb') as tar_output_file:
            total_num_bytes = 0
            last_console_update = 0
            start_time = time.time()
            printed_static_message = False
            # Pipe the output of tar into the output file
            while True:
                tar_exitcode = tar_subprocess.poll()
                data = tar_subprocess.stdout.read(4 * 1024 * 1024)
                if tar_exitcode is not None and len(data) == 0:
                    break
                tar_output_file.write(data)
                total_num_bytes += len(data)
                current_time = time.time()
                # Don't show status messages at all for very short tar
                # operations (< 1.0 sec)
                if current_time - last_console_update > 0.25 and current_time - start_time > 1.0:
                    if sys.stderr.isatty():
                        if last_console_update > 0:
                            sys.stderr.write("\r")
                        sys.stderr.write("Compressing target directory {dir}... ({kb_so_far:,} kb)".format(dir=src_dir, kb_so_far=total_num_bytes // 1024))
                        sys.stderr.flush()
                        last_console_update = current_time
                    elif not printed_static_message:
                        # Print a message (once only) when stderr is not
                        # going to a live console
                        sys.stderr.write("Compressing target directory %s..." % (src_dir,))
                        printed_static_message = True

        if last_console_update > 0:
            sys.stderr.write("\n")
        if tar_exitcode != 0:
            raise Exception("tar exited with non-zero exit code " + str(tar_exitcode))

        dxpy.set_workspace_id(build_project_id)

        remote_file = dxpy.upload_local_file(app_tarball_file, media_type="application/gzip",
                                             wait_on_close=True, show_progress=True)

        try:
            input_hash = {
                "input_file": dxpy.dxlink(remote_file),
                "build_options": build_options
                }
            if mode == 'app':
                input_hash["publish"] = publish
            api_options = {
                "name": "Remote build of %s" % (os.path.basename(src_dir),),
                "input": input_hash,
                "project": build_project_id,
                }
            if dest_folder:
                api_options["folder"] = dest_folder
            app_run_result = dxpy.api.app_run(builder_app, input_params=api_options)
            job_id = app_run_result["id"]
            print("Started builder job %s" % (job_id,))
            if watch:
                try:
                    subprocess.check_call(["dx", "watch", job_id])
                except subprocess.CalledProcessError as e:
                    if e.returncode == 3:
                        # Some kind of failure to build the app. The reason
                        # for the failure is probably self-evident from the
                        # job log (and if it's not, the CalledProcessError
                        # is not informative anyway), so just propagate the
                        # return code without additional remarks.
                        sys.exit(3)
                    else:
                        raise e

            dxpy.DXJob(job_id).wait_on_done(interval=1)

            if mode == 'applet':
                applet_id, _ = dxpy.get_dxlink_ids(dxpy.api.job_describe(job_id)['output']['output_applet'])
                return applet_id
            else:
                # TODO: determine and return the app ID, to allow
                # running the app if args.run is specified
                return None
        finally:
            if not using_temp_project_for_remote_build:
                dxpy.DXProject(build_project_id).remove_objects([remote_file.get_id()])
    finally:
        if using_temp_project_for_remote_build:
            dxpy.api.project_destroy(build_project_id, {"terminateJobs": True})
        shutil.rmtree(temp_dir)


def build_app_from(applet_desc, version, publish=False, do_try_update=True, bill_to_override=None,
                   return_object_dump=False, confirm=True, brief=False, **kwargs):

    applet_id = applet_desc["id"]
    app_name = applet_desc["name"]
    dxpy.executable_builder.verify_developer_rights('app-' + app_name)
    if not brief:
        logger.info("Will create app from the applet: %s (%s)" % (applet_desc["name"], applet_desc['id'],))

    applet_region = dxpy.api.project_describe(applet_desc["project"],
                                              input_params={"fields": {"region": True}})["region"]

    #TODO: make it possible to build multi region app by uploadling
    # the applet tarball to different regions
    regional_options = {
        applet_region: {'applet': applet_id}
    }

    # Certain metadata is not copied from an applet to the app
    # It must be passed explicitly otherwise default values will be
    # set by the API server or an error will be throw during app build
    # for required non-empty fields
    fields_to_inherit = (
        "summary", "title", "description", "developerNotes",
        "details", "access", "ignoreReuse"
    )
    inherited_metadata = {}
    for field in fields_to_inherit:
        if field in applet_desc:
            inherited_metadata[field] = applet_desc[field]

    required_non_empty = ("summary", "title", "description", "developerNotes")
    for field in required_non_empty:
        if field not in inherited_metadata or not inherited_metadata[field]:
            inherited_metadata[field] = applet_desc["name"]

    app_id = dxpy.app_builder.create_app_multi_region(regional_options,
                                                      app_name,
                                                      None,
                                                      publish=publish,
                                                      set_default=publish,
                                                      billTo=bill_to_override,
                                                      try_versions=version,
                                                      try_update=do_try_update,
                                                      confirm=confirm,
                                                      inherited_metadata=inherited_metadata,
                                                      brief=brief)
    app_describe = dxpy.api.app_describe(app_id)

    if not brief:
        if publish:
            logger.info("Uploaded and published app %s/%s (%s) successfully" % (app_describe["name"], app_describe["version"], app_id))
        else:
            logger.info("Uploaded app %s/%s (%s) successfully" % (app_describe["name"], app_describe["version"], app_id))
            logger.info("You can publish this app with: dx publish {n}/{v}".format(n=app_describe["name"], v=app_describe["version"]))

    return app_describe if return_object_dump else {"id": app_id}


def build_and_upload_locally(src_dir, mode, overwrite=False, archive=False, publish=False, destination_override=None,
                             version_override=None, bill_to_override=None, use_temp_build_project=True,
                             do_parallel_build=True, do_version_autonumbering=True, do_try_update=True,
                             do_check_syntax=True, dry_run=False,
                             return_object_dump=False, confirm=True, ensure_upload=False, force_symlinks=False,
                             region=None, brief=False, resources_dir=None, worker_resources_subpath="", **kwargs):
    dxpy.app_builder.build(src_dir, parallel_build=do_parallel_build)
    app_json = _parse_app_spec(src_dir)

    _check_suggestions(app_json, publish=publish)
    _verify_app_source_dir(src_dir, mode, enforce=do_check_syntax)
    if mode == "app" and not dry_run:
        dxpy.executable_builder.verify_developer_rights('app-' + app_json['name'])

    working_project = None
    using_temp_project = False
    override_folder = None
    override_applet_name = None

    enabled_regions = dxpy.app_builder.get_enabled_regions(app_json, region)

    # Cannot build multi-region app if `use_temp_build_project` is false.
    if enabled_regions is not None and len(enabled_regions) > 1 and not use_temp_build_project:
        raise dxpy.app_builder.AppBuilderException("Cannot specify --no-temp-build-project when building multi-region apps")

    # Prepare projects in which the app's underlying applets will be built (one per region).
    projects_by_region = None
    if mode == "applet" and destination_override:
        working_project, override_folder, override_applet_name = parse_destination(destination_override)
        if kwargs.get("name"):
            if override_applet_name:
                logger.warning("Name of the applet is set in both destination and extra args! "
                               "\"{}\" will be used!".format(kwargs.get("name")))
            override_applet_name = kwargs.get("name")

        region = dxpy.api.project_describe(working_project,
                                           input_params={"fields": {"region": True}})["region"]
        projects_by_region = {region: working_project}
    elif mode == "app" and use_temp_build_project and not dry_run:
        projects_by_region = {}
        if enabled_regions is not None:
            # Create temporary projects in each enabled region.
            try:
                for region in enabled_regions:
                    project_input = {
                        "name": "Temporary build project for dx-build-app in {r}".format(r=region),
                        "region": region
                    }
                    if bill_to_override:
                        project_input["billTo"] = bill_to_override
                    working_project = dxpy.api.project_new(project_input)["id"]
                    projects_by_region[region] = working_project
                    logger.debug("Created temporary project %s to build in" % (working_project,))
            except:
                # A /project/new request may fail if the requesting user is
                # not authorized to create projects in a certain region.
                dxpy.executable_builder.delete_temporary_projects(list(projects_by_region.values()))
                err_exit()
        else:
            # Create a temp project
            try:
                project_input = {"name": "Temporary build project for dx-build-app"}
                if bill_to_override:
                    project_input["billTo"] = bill_to_override
                working_project = dxpy.api.project_new(project_input)["id"]
            except:
                err_exit()
            region = dxpy.api.project_describe(working_project,
                                               input_params={"fields": {"region": True}})["region"]
            projects_by_region[region] = working_project
            logger.debug("Created temporary project %s to build in" % (working_project,))

        using_temp_project = True
    elif mode == "app" and not dry_run:
        # If we are not using temporary project(s) to build the executable,
        # then we should have a project context somewhere.
        try:
            project = app_json.get("project", dxpy.WORKSPACE_ID)
            region = dxpy.api.project_describe(project,
                                               input_params={"fields": {"region": True}})["region"]
        except Exception:
            err_exit()
        projects_by_region = {region: project}

    try:
        if mode == "applet" and working_project is None and dxpy.WORKSPACE_ID is None:
            parser.error("Can't create an applet without specifying a destination project; please use the -d/--destination flag to explicitly specify a project")

        if mode == "applet":
            dest_project = working_project or dxpy.WORKSPACE_ID or app_json.get("project", False)
            try:
                region = dxpy.api.project_describe(dest_project,
                                                   input_params={"fields": {"region": True}})["region"]
            except Exception:
                err_exit()
            projects_by_region = {region: dest_project}

            if not overwrite and not archive:
                # If we cannot overwrite or archive an existing applet and an
                # applet in the destination exists with the same name as this
                # one, then we should err out *before* uploading resources.
                try:
                    dest_name = override_applet_name or app_json.get('name') or os.path.basename(os.path.abspath(src_dir))
                except:
                    raise dxpy.app_builder.AppBuilderException("Could not determine applet name from specification + "
                                                               "(dxapp.json) or from working directory (%r)" % (src_dir,))
                dest_folder = override_folder or app_json.get('folder') or '/'
                if not dest_folder.endswith('/'):
                    dest_folder = dest_folder + '/'
                for result in dxpy.find_data_objects(classname="applet", name=dest_name, folder=dest_folder,
                                                     project=dest_project, recurse=False):
                    dest_path = dest_folder + dest_name
                    msg = "An applet already exists at {} (id {}) and neither".format(dest_path, result["id"])
                    msg += " -f/--overwrite nor -a/--archive were given."
                    raise dxpy.app_builder.AppBuilderException(msg)

        if dry_run:
            # Set a dummy "projects_by_region" so that we can exercise the dry
            # run flows for uploading resources bundles and applets below.
            projects_by_region = {"dummy-cloud:dummy-region": "project-dummy"}

        if projects_by_region is None:
            raise AssertionError("'projects_by_region' should not be None at this point")

        # "resources" can be used only with an app enabled in a single region and when
        # "regionalOptions" field is not specified.
        if "resources" in app_json and ("regionalOptions" in app_json or len(projects_by_region) > 1):
            error_message = "dxapp.json cannot contain a top-level \"resources\" field "
            error_message += "when the \"regionalOptions\" field is used or when "
            error_message += "the app is enabled in multiple regions"
            raise dxpy.app_builder.AppBuilderException(error_message)

        resources_bundles_by_region = {}
        for region, project in list(projects_by_region.items()):
            resources_bundles_by_region[region] = dxpy.app_builder.upload_resources(
                src_dir,
                project=project,
                folder=override_folder,
                ensure_upload=ensure_upload,
                force_symlinks=force_symlinks,
                brief=brief,
                resources_dir=resources_dir,
                worker_resources_subpath=worker_resources_subpath) if not dry_run else []

        # TODO: Clean up these applets if the app build fails.
        applet_ids_by_region = {}
        try:
            for region, project in list(projects_by_region.items()):
                applet_id, applet_spec = dxpy.app_builder.upload_applet(
                    src_dir,
                    resources_bundles_by_region[region],
                    check_name_collisions=(mode == "applet"),
                    overwrite=overwrite and mode == "applet",
                    archive=archive and mode == "applet",
                    project=project,
                    override_folder=override_folder,
                    override_name=override_applet_name,
                    dry_run=dry_run,
                    brief=brief,
                    **kwargs)
                if not dry_run:
                    logger.debug("Created applet " + applet_id + " successfully")
                applet_ids_by_region[region] = applet_id
        except:
            # Avoid leaking any bundled_resources files we may have
            # created, if applet creation fails. Note that if
            # using_temp_project, the entire project gets destroyed at
            # the end, so we don't bother.
            if not using_temp_project:
                for region, project in list(projects_by_region.items()):
                    objects_to_delete = [dxpy.get_dxlink_ids(bundled_resource_obj['id'])[0] for bundled_resource_obj in resources_bundles_by_region[region]]
                    if objects_to_delete:
                        dxpy.api.project_remove_objects(
                            dxpy.app_builder.get_destination_project(src_dir, project=project),
                            input_params={"objects": objects_to_delete})
            raise

        if dry_run:
            return

        applet_name = applet_spec['name']

        if mode == "app":
            if 'version' not in app_json:
                parser.error("dxapp.json contains no \"version\" field, but it is required to build an app")
            version = app_json['version']
            try_versions = [version_override or version]
            if not version_override and do_version_autonumbering:
                try_versions.append(version + _get_version_suffix(src_dir, version))

            additional_resources_by_region = {}
            if "regionalOptions" in app_json:
                for region, region_config in list(app_json["regionalOptions"].items()):
                    if "resources" in region_config:
                        additional_resources_by_region[region] = region_config["resources"]
            elif "resources" in app_json:
                additional_resources_by_region[list(projects_by_region.keys())[0]] = app_json["resources"]

            regional_options = {}
            for region in projects_by_region:
                regional_options[region] = {"applet": applet_ids_by_region[region]}
                if region in additional_resources_by_region:
                    regional_options[region]["resources"] = additional_resources_by_region[region]

            # add pricingPolicy separately for better readability
            if "regionalOptions" in app_json:
                for region, region_config in list(app_json["regionalOptions"].items()):
                    if "pricingPolicy" in region_config:
                        regional_options[region]["pricingPolicy"] = region_config["pricingPolicy"]

            app_id = dxpy.app_builder.create_app_multi_region(regional_options,
                                                              applet_name,
                                                              src_dir,
                                                              publish=publish,
                                                              set_default=publish,
                                                              billTo=bill_to_override,
                                                              try_versions=try_versions,
                                                              try_update=do_try_update,
                                                              confirm=confirm,
                                                              brief=brief)

            app_describe = dxpy.api.app_describe(app_id)

            if not brief:
                if publish:
                    logger.info("Uploaded and published app %s/%s (%s) successfully" % (app_describe["name"], app_describe["version"], app_id))
                else:
                    logger.info("Uploaded app %s/%s (%s) successfully" % (app_describe["name"], app_describe["version"], app_id))
                    logger.info("You can publish this app with: dx publish {n}/{v}".format(n=app_describe["name"], v=app_describe["version"]))

            return app_describe if return_object_dump else {"id": app_id}

        elif mode == "applet":
            return dxpy.api.applet_describe(applet_id) if return_object_dump else {"id": applet_id}
        else:
            raise dxpy.app_builder.AppBuilderException("Unrecognized mode %r" % (mode,))

    finally:
        # Clean up after ourselves.
        if using_temp_project:
            dxpy.executable_builder.delete_temporary_projects(list(projects_by_region.values()))

def get_destination_region(destination):
    """
    :param destination: The destination path for building the applet, as given by the --destination option to "dx build". Will be in the form [PROJECT_NAME_OR_ID:][/[FOLDER/][NAME]].
    :type destination: str

    :returns: The name of the region in which the applet will be built, e.g. 'aws:us-east-1'. It doesn't take into account the destination project specified in dxapp.json.
    :rtype: str
    """
    if destination:
        dest_project_id, _, _ = parse_destination(destination)
    else:
        dest_project_id = dxpy.WORKSPACE_ID
    return dxpy.api.project_describe(dest_project_id, input_params={"fields": {"region": True}})["region"]


def get_project_to_check(destination, extra_args):
    # extra args overrides the destination argument
    # so we're checking it first
    if "project" in extra_args:
        return extra_args["project"]
    if destination:
        dest_project_id, _, _ = parse_destination(destination)
        # checkFeatureAccess is not implemented on the container
        if dest_project_id.startswith("container-"):
            dest_project_id = dxpy.PROJECT_CONTEXT_ID
        return dest_project_id
    else:
        return dxpy.PROJECT_CONTEXT_ID


def verify_nf_license(destination, extra_args):
    dest_project_to_check = get_project_to_check(destination, extra_args)
    features = dxpy.DXHTTPRequest("/" + dest_project_to_check + "/checkFeatureAccess", {"features": ["dxNextflow"]}, always_retry=True).get("features", {})
    dx_nextflow_lic = features.get("dxNextflow", False)
    if not dx_nextflow_lic:
        raise dxpy.app_builder.AppBuilderException("PermissionDenied: billTo of the applet's destination project must have the dxNextflow feature enabled. For inquiries, please contact support@dnanexus.com")

def _build_app(args, extra_args):
    """Builds an app or applet and returns the resulting executable ID
    (unless it was a dry-run, in which case None is returned).

    TODO: remote app builds still return None, but we should fix this.

    """
    resources_dir = None
    source_dir = args.src_dir
    worker_resources_subpath = ""  # no subpath, files will be saved to root directory by default.

    if args.nextflow:
        verify_nf_license(args.destination, extra_args)

    # determine if a nextflow applet ought to be built with Nextflow Pipeline Importer (NPI) app
    build_nf_with_npi = any([x for x in [args.repository, args.cache_docker]])
    # this is to ensure to not call any more NPI executions if already inside an NPI job.
    build_nf_with_npi = False if is_importer_job() else build_nf_with_npi

    if args.nextflow and (not build_nf_with_npi):
        source_dir = prepare_nextflow(
            resources_dir=args.src_dir,
            profile=args.profile,
            region=get_destination_region(args.destination),
            cache_docker=args.cache_docker,
            nextflow_pipeline_params=args.nextflow_pipeline_params
        )
        resources_dir = args.src_dir
        worker_resources_subpath = get_resources_subpath(resources_dir)
    if args._from:
        # BUILD FROM EXISTING APPLET
        output = build_app_from(
            args._from,
            [args.version_override],
            publish=args.publish,
            do_try_update=args.update,
            bill_to_override=args.bill_to,
            confirm=args.confirm,
            return_object_dump=args.json,
            brief=args.brief,
            **extra_args
        )
        if output is not None and args.run is None:
            print(json.dumps(output))

        return output['id']

    if not args.remote and not build_nf_with_npi:  # building with NF repository or cache_docker is done remotely by npi
        # LOCAL BUILD
        output = build_and_upload_locally(
            source_dir,
            args.mode,
            overwrite=args.overwrite,
            archive=args.archive,
            publish=args.publish,
            destination_override=args.destination,
            version_override=args.version_override,
            bill_to_override=args.bill_to,
            use_temp_build_project=args.use_temp_build_project,
            do_parallel_build=args.parallel_build,
            do_version_autonumbering=args.version_autonumbering,
            do_try_update=args.update,
            do_check_syntax=args.check_syntax,
            ensure_upload=args.ensure_upload,
            force_symlinks=args.force_symlinks,
            dry_run=args.dry_run,
            confirm=args.confirm,
            return_object_dump=args.json,
            region=args.region,
            brief=args.brief,
            resources_dir=resources_dir,
            worker_resources_subpath=worker_resources_subpath,
            **extra_args
            )

        if output is not None and args.run is None:
            print(json.dumps(output))

        if args.dry_run:
            return None

        return output['id']

    else:
        # REMOTE BUILD
        # The following flags might be useful in conjunction with
        # --remote. To enable these, we need to learn how to pass these
        # options through to the interior call of dx_build_app(let).
        incompatible_options = "--cache-docker" if args.cache_docker else "--remote and --repository"
        if args.dry_run:
            parser.error('{} cannot be combined with --dry-run'.format(incompatible_options))
        if args.overwrite:
            parser.error('{} cannot be combined with --overwrite/-f'.format(incompatible_options))
        if args.archive:
            parser.error('{} cannot be combined with --archive/-a'.format(incompatible_options))

        # The following flags are probably not useful in conjunction
        # with --remote.
        if args.json:
            parser.error('{} cannot be combined with --json'.format(incompatible_options))
        if not args.use_temp_build_project:
            parser.error('{} cannot be combined with --no-temp-build-project'.format(incompatible_options))

        if isinstance(args.region, list) and len(args.region) > 1:
            parser.error('--region can only be specified once for remote builds')
        region = args.region[0] if args.region is not None else None

        more_kwargs = {}
        if args.version_override:
            more_kwargs['version_override'] = args.version_override
        if args.bill_to:
            more_kwargs['bill_to_override'] = args.bill_to
        if not args.version_autonumbering:
            more_kwargs['do_version_autonumbering'] = False
        if not args.update:
            more_kwargs['do_try_update'] = False
        if not args.parallel_build:
            more_kwargs['do_parallel_build'] = False
        if not args.check_syntax:
            more_kwargs['do_check_syntax'] = False


        if args.nextflow and build_nf_with_npi:
            nf_scr = args.repository
            if (not args.repository) and args.src_dir:
                logger.info(
                    "Building nextflow pipeline with the Nextflow Pipeline Importer app. "
                    "Uploading the local nextflow source to the platform"
                )
                dest_project = get_project_to_check(args.destination, extra_args)
                _, dest_folder, _ = parse_destination(args.destination)
                dest_folder = dest_folder or ""
                upload_destination_dir = os.path.join(
                    dest_folder, ".nf_source"
                ).strip("/")
                qualified_upload_dest = ":".join([dest_project, "/" + upload_destination_dir + "/"])
                dest_folder_exists = False

                try:
                    dest_folder_exists = check_folder_exists(
                        project=dest_project,
                        path="/" + os.path.join(dest_folder, ".nf_source").strip("/"),
                        folder_name=os.path.basename(args.src_dir)
                    )
                except ResolutionError:
                    logger.info(
                        "Destination folder {} does not exist. Creating and uploading the pipeline source.".format(
                            qualified_upload_dest
                        )
                    )
                if dest_folder_exists:
                    raise dxpy.app_builder.AppBuilderException(
                        "Folder {} exists in the project {}. Remove the directory to avoid file duplication and retry".format(
                            os.path.join(upload_destination_dir, os.path.basename(args.src_dir)), dest_project
                        )
                    )
                else:
                    upload_cmd = ["dx", "upload", args.src_dir, "-r", "-o", qualified_upload_dest, "-p"]
                    _ = subprocess.check_output(upload_cmd)
                    nf_scr = os.path.join(qualified_upload_dest, os.path.basename(args.src_dir))
            return build_pipeline_with_npi(
                repository=nf_scr,
                tag=args.tag,
                cache_docker=args.cache_docker,
                docker_secrets=args.docker_secrets,
                nextflow_pipeline_params=args.nextflow_pipeline_params,
                profile=args.profile,
                git_creds=args.git_credentials,
                brief=args.brief,
                destination=args.destination,
                extra_args=extra_args
            )
        app_json = _parse_app_spec(source_dir)
        _check_suggestions(app_json, publish=args.publish)
        _verify_app_source_dir(source_dir, args.mode)
        if args.mode == "app" and not args.dry_run:
            dxpy.executable_builder.verify_developer_rights('app-' + app_json['name'])

        return _build_app_remote(args.mode, source_dir, destination_override=args.destination,
                                 publish=args.publish,
                                 region=region, watch=args.watch, **more_kwargs)


def build(args):
    try:
        process_extra_args(args)
        executable_id = _build_app(args, args.extra_args or {})
    except dxpy.app_builder.AppBuilderException as e:
        # AppBuilderException represents errors during app building
        # that could reasonably have been anticipated by the user.
        print("Error: %s" % (e.args,), file=sys.stderr)
        sys.exit(3)
    except dxpy.exceptions.DXAPIError as e:
        print("Error: %s" % (e,), file=sys.stderr)
        sys.exit(3)
    if args.run is not None:
        if executable_id is None:
            raise AssertionError('Expected executable_id to be set here')

        try:
            subprocess.check_call(['dx', 'run', executable_id, '--priority', 'high'] + args.run)
        except subprocess.CalledProcessError as e:
            sys.exit(e.returncode)
        except:
            err_exit()


def main(**kwargs):
    """
    Entry point for dx-build-app(let).

    Don't call this function as a subroutine in your program! It is liable to
    sys.exit your program when it detects certain error conditions, so you
    can't recover from those as you could if it raised exceptions. Instead,
    call dx_build_app.build_and_upload_locally which provides the real
    implementation for dx-build-app(let) but is easier to use in your program.
    """

    if len(sys.argv) > 0:
        if sys.argv[0].endswith('dx-build-app'):
            logging.warn('Warning: dx-build-app has been replaced with "dx build --create-app". Please update your scripts.')
        elif sys.argv[0].endswith('dx-build-applet'):
            logging.warn('Warning: dx-build-applet has been replaced with "dx build". Please update your scripts.')
        exit(0)

if __name__ == '__main__':
    main()



# Copyright (C) 2013-2016 DNAnexus, Inc.
#
# This file is part of dx-toolkit (DNAnexus platform client libraries).
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may not
#   use this file except in compliance with the License. You may obtain a copy
#   of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from __future__ import print_function, unicode_literals, division, absolute_import

import sys, locale, threading, hashlib

try:
    sys_encoding = locale.getdefaultlocale()[1] or "UTF-8"
except Exception:
    sys_encoding = "UTF-8"

USING_PYTHON2 = True if sys.version_info < (3, 0) else False

basestring = (str, bytes)

THREAD_TIMEOUT_MAX = threading.TIMEOUT_MAX

# Support FIPS enabled Python
def md5_hasher():
    try:
        md5_hasher = hashlib.new('md5', usedforsecurity=False)
    except:
        md5_hasher = hashlib.new('md5')
    return md5_hasher
